/**
 * ISS Copyright 2020
 * @author Calin Bira (calin.bira@spacescience.ro / calin.bira@upb.ro)
 */

#pragma once

#include <Adafruit_GFX_AS.h>
#include <Adafruit_ILI9341.h>
#include "Bitmap.hpp"

class issDisplay {

  #define TFT_DC 3
  #define TFT_CS 4
  #define TFT_LED 2

  #define BACKLIGHT_ON LOW
  #define BACKLIGHT_OFF HIGH
      
  uint8_t mIndex; // next coordinate where we may write on the display
  Adafruit_ILI9341 *ptft;
  const uint8_t ROTATION = LANDSCAPE;
  
  //font size is 2 or 4
  #define FONT_SIZE 1
  #define ROW_HEIGHT (6.7 * FONT_SIZE)
  const uint8_t MAX_ROW_INDEX[4] = {48 / FONT_SIZE, 36 / FONT_SIZE, 48 / FONT_SIZE, 36 / FONT_SIZE};

  uint8_t backLight;
  uint8_t printDebugInfoPin = 0;
  
  public:
    enum Rotation {
       PORTRAIT           = 0,
       LANDSCAPE_INVERTED = 1,
       PORTRAIT_INVERTED  = 2,
       LANDSCAPE          = 3
    };


    void setBacklight(uint8_t val) {
      backLight = val;
      digitalWrite(TFT_LED, val);
    }

    void fadeBacklight(uint8_t val) {
      if ((val == BACKLIGHT_OFF) && (backLight == BACKLIGHT_ON)) {
        for(int i=0; i<=255; i++){ analogWrite(TFT_LED, i); delay(10);}
        digitalWrite(TFT_LED, BACKLIGHT_OFF);
        backLight = BACKLIGHT_OFF;
      }
      else if ((val == BACKLIGHT_ON) && (backLight == BACKLIGHT_OFF)) {
        for(int i=255; i>=0; i--){ analogWrite(TFT_LED, i); delay(10);}
        digitalWrite(TFT_LED, BACKLIGHT_ON);
        backLight = BACKLIGHT_ON;
      }
    }
    
    void init() {
      mIndex = 0;     
      // Use hardware SPI (on Uno, #13, #12, #11) and the above for CS/DC
      ptft = new Adafruit_ILI9341(TFT_CS, TFT_DC, 8, 9, 6, 10);
            
      ptft->begin();
      ptft->setRotation(ROTATION);
      ptft->setTextColor(ILI9341_BLACK, ILI9341_WHITE);
     
      pinMode(TFT_LED, OUTPUT);
      setBacklight(BACKLIGHT_OFF);
            
      long startms = millis();
        drawLogoFast();
        //drawLogoSlow();
      long endms = millis();
      drawString(String(endms - startms) + " ms to draw logo");
      fadeBacklight(BACKLIGHT_ON);
        
      //printDebugInfoPin = _printDebugInfoPin;      
      //pinMode(printDebugInfoPin, INPUT_PULLUP);
    }

    uint8_t shouldPrintText() {
      return true; // (digitalRead(printDebugInfoPin) == 0);
    }

    void drawString(char*msg, int posX, int posY, int sizeZ) {
      if (!shouldPrintText()) return;

      ptft->setTextSize(sizeZ);
      ptft->setCursor(posX, posY);
      ptft->println(msg);
    }

    void drawNumber(int msg, int posX, int posY, int sizeZ) {
      if (!shouldPrintText()) return;

      ptft->setTextSize(sizeZ);
      ptft->setCursor(posX, posY);
      ptft->println(msg);
    }

/*
    void drawFloat(float msg, int decimals, int posX, int posY, int sizeZ) {
      if (!shouldPrintText()) return;
      ptft->drawFloat(msg, decimals, posX, posY, sizeZ);
    }
*/

    void setY(uint16_t val) {
        mIndex = val / ROW_HEIGHT;
    }

    void drawNumber(int msg) {
      if (!shouldPrintText()) return; 
      
      ptft->setTextSize(getFontSize());
      ptft->setCursor(0, getNextY());
      ptft->println(msg);
    }

    void drawString(String msg, int posX, int posY, int sizeZ) {
      if (!shouldPrintText()) return;
      
      msg.replace('\n', '_');
      msg.replace('\r', '_');

      ptft->setTextSize(sizeZ);
      ptft->setCursor(posX, posY);
      ptft->println((char*)msg.c_str());
    }

    void drawString(String msg) {
      if (!shouldPrintText()) return;
      
      if (msg.length() <= 40)
        drawString(String (millis()/1000) + " L" + String(msg.length()) + " " + msg, 0, getNextY(), getFontSize());
      else {
        drawString("Line too long, got split: ", 0, getNextY(), getFontSize());
        int from_offset = 0;
        while (from_offset < msg.length()) {
          int to_offset = from_offset + 40;
          if (to_offset > msg.length()) to_offset = msg.length();
          
          drawString(msg.substring(from_offset, to_offset + 40), 0, getNextY(), getFontSize());
          from_offset = to_offset;
        }
      }
    }
    
    uint16_t getNextY() {
      uint16_t currentY = mIndex * ROW_HEIGHT;
      mIndex++;
      if (mIndex > MAX_ROW_INDEX[ROTATION]) {
        mIndex = 0;      // set offset to begining of screen
        ptft->fillScreen(ILI9341_BLACK); // erase screen
      }
      return currentY;
    }

    static inline uint8_t getFontSize() {
      return FONT_SIZE;
    }
    
    // const uint8_t BitmapPalette[256*3] PROGMEM = {256 items of bgr}
    // const uint8_t BitmapPixels[19200] PROGMEM = { 2bpp }
    // void Adafruit_ILI9341_AS::drawPixel(int16_t x, int16_t y, uint16_t color)

    static uint16_t indexToColor(uint8_t colorindex) {
      //bgr 888 to 565 conversion
      uint16_t b = pgm_read_byte_near(BitmapPalette + 3 * colorindex) >> 3;
      uint16_t g = pgm_read_byte_near(BitmapPalette + 3 * colorindex + 1) >> 2;
      uint16_t r = pgm_read_byte_near(BitmapPalette + 3 * colorindex + 2) >> 3;
      
      return (r << 11) + (g << 5) + b;
    }

    void drawLogoSlow() {
      for (uint16_t y = 0; y < BitmapHeight; y++)
      for (uint16_t x = 0; x < BitmapWidth; x+=4) {
        uint8_t pixel = pgm_read_byte_near(BitmapPixels + y * (BitmapWidth / 4) + (x >> 2));
        ptft->drawPixel(x, y, indexToColor(pixel >> 6));
        ptft->drawPixel(x+1, y, indexToColor((pixel >> 4) & 0x03));
        ptft->drawPixel(x+2, y, indexToColor((pixel >> 2) & 0x03));
        ptft->drawPixel(x+3, y, indexToColor(pixel & 0x03));
      }
    }
    
    void drawLogoFast() {
      uint16_t x = 0;
      uint16_t oldx = 0;
      uint8_t old_pixel = pgm_read_byte_near(BitmapPixels);
      uint8_t new_pixel = old_pixel;
      
      for (uint16_t y = 0; y < BitmapHeight; ) {
          do {
                x += 4;
                new_pixel = pgm_read_byte_near(BitmapPixels + y * (BitmapWidth / 4) + (x >> 2));                
          }
          while (((new_pixel == 0x00) || (new_pixel == 0x55) || (new_pixel == 0xAA) || (new_pixel == 0xFF)) 
              && (new_pixel == old_pixel) && (x < BitmapWidth));
          
          if (x - oldx == 4) {
            ptft->drawPixel(oldx, y, indexToColor(old_pixel >> 6));
            ptft->drawPixel(oldx + 1, y, indexToColor((old_pixel >> 4) & 0x03));
            ptft->drawPixel(oldx + 2, y, indexToColor((old_pixel >> 2) & 0x03));
            ptft->drawPixel(oldx + 3, y, indexToColor(old_pixel & 0x03));
          }
          else
            ptft->drawFastHLine(oldx, y, x - oldx, indexToColor(old_pixel & 0x03));
          
          if (x == BitmapWidth) {
            x = 0;
            y++;
            
            oldx = 0;
            old_pixel = pgm_read_byte_near(BitmapPixels + y * (BitmapWidth / 4) + (x >> 2));
          } else {
            oldx = x;
            old_pixel = new_pixel;
          }          
        }
    }

    enum {
      LOGO_SCREEN_TEMPERETURE_ROW   = 10,
      LOGO_SCREEN_SPIROMETER_ROW,
      LOGO_SCREEN_SPO2_ROW,
      LOGO_SCREEN_BLOOD_PRESSURE_ROW,
      LOGO_SCREEN_GLUCOMETER_ROW,
      LOGO_SCREEN_BODY_POSITION_ROW,
      LOGO_SCREEN_BODY_SCALE_ROW,
    }LogoScreenSensorRow_t; 

    unsigned short getForcedRowX() {
      return ROW_HEIGHT / 2;
    }
    unsigned short getForcedRowY(unsigned short row) {
      return ROW_HEIGHT * row + ROW_HEIGHT / 2; 
      // add an offset of half-row for better separation between logo and bottom screen
    }
    
    void drawStringForced(String msg, int posX, int posY) {
      msg.replace('\n', '_');
      msg.replace('\r', '_');
      
      ptft->setTextSize(getFontSize());
      ptft->setCursor(posX, posY);
      ptft->println((char*)msg.c_str());
    }
    
    ~issDisplay() {
      delete(ptft);
    }
};
