/*
 *  Class for ISS ChoiceReactionTime (CRT) module
 *  @author Costin Vasile (costin.vasile1003@upb.ro)
 *  @author Calin Bira (calin.bira@upb.ro)
 *  
 *  This class implements a CRT test as follows:
 *    if (CRT was ordered) {
 *        while (experimentTime is has not elapsed) {
 *          lite-up a R,G or B led, randomly (keep it bright for 200ms)
 *          check if correct button was pressed (RIGHT for B, LEFT for R or G)
 *          check if wrong button was pressed
 *          wait for SOA time to pass (typ 2500 ms) *        
 *        }      
 *        connect to wifi
 *        connect to cloud-server
 *        data to cloud-server    
 *    }*  
 *  
 */

#pragma once

//#include "ArduinoLowPower.h"
#include "issCRTStats.hpp"
#include "issButton.hpp"
#include "issUtils.hpp"

/**
* Enum for the random-generated next-light
*/
enum CRT_LED_NUM {
    CRT_LED_START = 0,
    CRT_LED_G = 0,
    CRT_LED_R = 1,
    CRT_LED_B = 2,
    CRT_LED_STOP = 3
};
    
class issCRT2 {
  /**
   * Enum for the ISR-driven FSM state
   */
    enum CRTstate {
      CRT_IDLE,
      CRT_SOUND_STARTED,
      CRT_SOUND_ENDED,
      CRT_ROUND_STARTED,
      CRT_ROUND_WAIT_INIT,
      CRT_ROUND_INIT,
      CRT_ROUND_CHECK,
      CRT_BUTTON_DELAY,
      CRT_ROUND_CHECKED,
      CRT_ROUND_ENDED,
      CRT_JUST_FINISHED,
      CRT_FINISHED_BLUE,
      CRT_FINISHED_ALL,
      CRT_DATA_SENT
    };

    // Arduino GPIOs
    int mCRT_BUTTON_LEFT; /**< pin of LEFT button */ 
    int mCRT_BUTTON_RIGHT;/**< pin of RIGHT button */

    int mEQUID; /**< Equipment ID, used when sending data to cloud */

    issCRTStats R_stats; /**< Red-light statistics */
    issCRTStats G_stats; /**< Green-light statistics */
    issCRTStats B_stats; /**< Blue-light statistics */
    
    CRTstate mState = CRT_IDLE; /**< current FSM state */

    issSmartRGB *pRGB; 
    issBattery* pBattery = NULL;
    issButton* pButtonLeft = NULL;
    issButton* pButtonRight = NULL;
    issRandom* pRandom = NULL;
    issSpeaker* pSpeaker = NULL;

    volatile long stagems;  /**< "current" millisecond timestamp, written by ISR */
    volatile long SOA_time_start; /**< SOA (round) start time millisecond timestamp, written by ISR */
    volatile long experimentStartTime; /**< Experiment start time millisecond timestamp, written by ISR */

  public:
    issCRT2(issSmartRGB *pLed, 
           issBattery *pBatt,
           issSpeaker *pSpeak,
            int crtButtonLeft, int crtButtonRight, 
            int randomPin, int equid) {
      pButtonLeft = new issButton(crtButtonLeft);
      pButtonRight = new issButton(crtButtonRight);
      pRandom = new issRandom(randomPin);

      mEQUID = equid;
      pRGB = pLed;
      pBattery = pBatt;
      pSpeaker = pSpeak;
    }

    void reset(){
        R_stats.resetStats();
        G_stats.resetStats();
        B_stats.resetStats();
        
        mState = CRT_IDLE;
    }

    void debugMode() {
      while (true) {
        if (pButtonLeft->isButtonPressed()) Serial.println("l");
        if (pButtonRight->isButtonPressed()) Serial.println("r");
        delay(50);
      }
    }
    
    void interruptLoop() {
      static uint8_t led;
      static uint8_t i = 0;
      static bool correctfirsttime;
      static long correcttime;
                  
      pButtonLeft->loopInterrupt();
      pButtonRight->loopInterrupt();
      
      switch (mState) {
        case CRT_IDLE:
            if (pButtonRight->getMillisContinuousPressed() > 2000) {
              //mState = CRT_STARTED_LED_R;
              mState = CRT_SOUND_STARTED;
              stagems = millis();
              i = 0;
            }
            break;
            
        case CRT_SOUND_STARTED:
            pRGB->blinkGreen(100, 50, true);
            if (millis() - stagems > 150) {
              stagems = millis();
              if (i++ < 10) {
                //
                pSpeaker->doCRTstart();
              }
              else {
                mState = CRT_SOUND_ENDED;                
              }
            }
            break;

        case CRT_SOUND_ENDED:
            pRGB->switchAllOff();
            if ((millis() - stagems) > CRT_DELAY_MS) {
              
              experimentStartTime = stagems;              
              mState = CRT_ROUND_WAIT_INIT;
            }
            break;

        case CRT_ROUND_WAIT_INIT:
            pRGB->switchAllOff();
            stagems = millis();
            SOA_time_start = millis();              
            
            mState = CRT_ROUND_INIT;
            break;
            
        case CRT_ROUND_INIT:
            led = pRandom->getRandomNumber(CRT_LED_START, CRT_LED_STOP);

            if      (led == CRT_LED_R) pRGB->switchRedOn();
            else if (led == CRT_LED_G) pRGB->switchGreenOn();
            else if (led == CRT_LED_B) pRGB->switchBlueOn();
            //Serial.println("CRT_ROUND_INIT: for led " + String(led));
            
            mState = CRT_ROUND_STARTED;
            break;
            
        case CRT_ROUND_STARTED:
            //Serial.println("CRT_ROUND_STARTED");
            correctfirsttime = true;
            
            mState = CRT_ROUND_CHECK;
            break;

        case CRT_ROUND_CHECK: {
            long checktime = millis();
            if (checktime - stagems >= 200) {
                pRGB->switchAllOff();
            }
        
            if (led == CRT_LED_B) {
              if (pButtonLeft->isButtonPressed()) {
                  correctfirsttime = 0;
                  B_stats.updateStats(false, checktime-stagems);
                  pSpeaker->doCRTwrong();
                  //keep state CRT_ROUND_CHECK
              } 
              else if (pButtonRight->isButtonPressed()) {
                mState = CRT_BUTTON_DELAY;
                correcttime = millis();
              }
            }
            else {
              if (pButtonRight->isButtonPressed()) {
                  correctfirsttime = 0;
                  if (led == CRT_LED_G)
                    G_stats.updateStats(false, checktime-stagems);
                  else
                    R_stats.updateStats(false, checktime-stagems);
                     
                  pSpeaker->doCRTwrong();
                  //keep state CRT_ROUND_CHECK
              }
              else if (pButtonLeft->isButtonPressed()) {
                mState = CRT_BUTTON_DELAY;
                correcttime = millis();
              }
            }
            
            break;
        }
        case CRT_BUTTON_DELAY: {
             if ((pButtonLeft->getMillisContinuousHigh() > CRT_BUTTON_DELAY_MS) && 
                  (pButtonRight->getMillisContinuousHigh() > CRT_BUTTON_DELAY_MS))
                mState = CRT_ROUND_CHECKED;
             break;
        }
        case CRT_ROUND_CHECKED: {
            //Serial.println("CRT_ROUND_CHECKED");
            //long stop = millis();
            //if (stop - stagems > 3) stop = stop - 3; //substract 3ms of the FSM: ROUND_INIT, ROUND_CHECK, ROUND_CHECKED
            pRGB->switchAllOff();
            // data interpretation
            if (led == CRT_LED_R) {
              R_stats.updateStats(true, correcttime-stagems);
              Serial.println(String("<CRT>") + String(correctfirsttime) + " R " + String(R_stats.getlastWrongTime()) +  " " + String(R_stats.getlastCorrectTime()));
              R_stats.advance();
            }
            else if (led == CRT_LED_G) {
              G_stats.updateStats(true, correcttime-stagems);
              Serial.println(String("<CRT>") + String(correctfirsttime) + " G " + String(G_stats.getlastWrongTime()) +  " " + String(G_stats.getlastCorrectTime()));
              G_stats.advance();
            }
            else if (led == CRT_LED_B) {
              B_stats.updateStats(true, correcttime-stagems);
              Serial.println(String("<CRT>") + String(correctfirsttime) + " B " + String(B_stats.getlastWrongTime()) +  " " + String(B_stats.getlastCorrectTime()));
              B_stats.advance();
            }

            mState = CRT_ROUND_ENDED;            
            break;
        }
        
        case CRT_ROUND_ENDED:
            if (millis() - stagems > CRT_AFTER__SOA_TIMEOUT_MS) {
              if (millis() - experimentStartTime > CRT_EXPERIMENT_TIME_SECONDS*1000) {
                if (R_stats.hasData() && G_stats.hasData() && B_stats.hasData()) {
                  mState = CRT_JUST_FINISHED; // stop experiment
                  //Serial.println("CRT_ROUND_ENDED stop experiment");
                }
                else {
                  mState = CRT_ROUND_WAIT_INIT; // go to one more round
                  //Serial.println("CRT_ROUND_ENDED CRT_ROUND_INIT - not enough data");
                }
              }
              else {
                mState = CRT_ROUND_WAIT_INIT; // go to next round
                //Serial.println("CRT_ROUND_ENDED CRT_ROUND_INIT - SOA");
              }
            }
            //else keep CRT_ROUND_ENDED
            break;

        case CRT_JUST_FINISHED: 
            Serial.println();
            Serial.println("\t####################### Results #######################");
            Serial.println("\tCOLOR\t|\tcorrect\t|\twrong\t|\tmax(ms)\t|\tmin(ms)\t|\tmean(ms)\t|\tdevstd(n/a)");
            Serial.print("\tRED:\t|\t"); R_stats.printlnStats();
            Serial.print("\tGREEN:\t|\t"); G_stats.printlnStats();
            Serial.print("\tBLUE:\t|\t"); B_stats.printlnStats();
            Serial.println("\t######################################################");
            Serial.println();
            pSpeaker->doCRTfinished();

            mState = CRT_FINISHED_BLUE;
            i = 0;
            stagems = millis();
            break;
            
        case CRT_FINISHED_BLUE:
            if (millis() - stagems > 150) {
              if (i++ < 10)
                 pRGB->blinkBlue(100, 50, true);
              else
                 mState = CRT_FINISHED_ALL;

              stagems = millis();
            }            
            break;
            
        case CRT_FINISHED_ALL:
            pRGB->switchAllOff();
            break; // wait outside interrupt to send the data
        
        default:
            break;
      }
      //Serial.println(mState);
    }

    int getStartTimestamp() {
      return experimentStartTime;
    }

    bool measFinished() {
      return mState == CRT_FINISHED_ALL;
    }
        
    String getStatsJSON(int duration_s){
      String data = "{";
      data += "\"NETWORK_DELAY\": " + String(duration_s);
      data += ", \"EQUID\": \"" + String(mEQUID) + "\"";
      data += ", \"CRT_R_STATS\":" + R_stats.getStats_asJSONLine();
      data += ", \"CRT_G_STATS\":" + G_stats.getStats_asJSONLine();
      data += ", \"CRT_B_STATS\":" + B_stats.getStats_asJSONLine();
      
      if (pBattery != NULL)
        data += ", \"CRT_BATTERY\":" + String(pBattery->getBatteryPercent());
      data += "}";

      return data;
    }
};

/*
class issCRT1 {
    // Arduino GPIOs
    int mCRT_BUTTON_LEFT;
    int mCRT_BUTTON_RIGHT;

    // Equipment ID, used when sending data to cloud
    int mEQUID;

    issCRTStats R_stats;
    issCRTStats G_stats;
    issCRTStats B_stats;

    bool mMeasFinished = true;
    bool mEnabled = false;
    
    issSmartRGB *pRGB;
    issBattery *pBattery = NULL;
    issButton *pButtonLeft = NULL;
    issButton *pButtonRight = NULL;

  public:
    issCRT1(issSmartRGB *pLed, 
           issBattery *pBatt,
            int crtButtonLeft, int crtButtonRight, int equid) {
      
      pButtonLeft = new issButton(crtButtonLeft);
      pButtonRight = new issButton(crtButtonRight);

      mEQUID = equid;
      pRGB = pLed;
      pBattery = pBatt;
    }

    void setup() {
      pRGB->blinkRed(800);
      pRGB->blinkGreen(800);
      pRGB->blinkBlue(800);

      R_stats.resetStats();
      G_stats.resetStats();
      B_stats.resetStats();

      //debugMode();
    }

    void debugMode() {
      while (true) {
        if (issButton::isButtonPressed(mCRT_BUTTON_RIGHT)) Serial.println("l");
        if (issButton::isButtonPressed(mCRT_BUTTON_LEFT)) Serial.println("r");
        delay(50);
      }
    }

    void runOnce() {
      static long ExperimentTime = millis();
      pRGB->switchAllOff();

      if (mMeasFinished) {
        R_stats.resetStats();
        G_stats.resetStats();
        B_stats.resetStats();

        mMeasFinished = false;
        Serial.println("\n\n############## Starting new CRT measurement ##############");
        delay(5000);
        ExperimentTime = millis();
      }

      long SOA_time = millis();

      int led = random(0, 3);

      long start = millis(); // measurement start
      if      (led == CRT_LED_R) pRGB->switchRedOn();
      else if (led == CRT_LED_G) pRGB->switchGreenOn();
      else if (led == CRT_LED_B) pRGB->switchBlueOn();
        
      // actual measurement
      bool correct = true;
      if (led == CRT_LED_B) // target
        while (!issButton::isButtonPressed(mCRT_BUTTON_RIGHT)) {
          if (issButton::isButtonPressed(mCRT_BUTTON_LEFT)) {
            correct = false;
            tone(BUZZER, WRONG_FREQ, WRONG_DUR);
          }
          if (millis() - start >= 200) pRGB->switchAllOff();
        }
      else
        while (!issButton::isButtonPressed(mCRT_BUTTON_LEFT)) {
          if (issButton::isButtonPressed(mCRT_BUTTON_RIGHT)) {
            correct = false;
            tone(BUZZER, WRONG_FREQ, WRONG_DUR);
          }
          if (millis() - start >= 200) pRGB->switchAllOff();
        }
      long stop = millis();
      pRGB->switchAllOff();

      // data interpretation
      if (led == CRT_LED_R) {
        Serial.println(String(correct) + " R " + String(stop - start));
        R_stats.updateStats(correct, stop-start);        
      }
      else if (led == CRT_LED_G) {
        Serial.println(String(correct) + " G " + String(stop - start));
        G_stats.updateStats(correct, stop-start);
      }
      else if (led == CRT_LED_B) {
        Serial.println(String(correct) + " B " + String(stop - start));
        B_stats.updateStats(correct, stop-start);
      }


      while (millis() - SOA_time < 2500) {
        //if (digitalRead(BUTTON_LEFT) == LOW) Serial.println("LEFT");
        //if (digitalRead(BUTTON_RIGHT) == LOW) Serial.println("RIGHT");
      }
      ;//wait for SOA

      if (millis() - ExperimentTime > CRT_EXPERIMENT_TIME_SECONDS * 1000UL) { // X seconds
        Serial.println();
        Serial.println("####################### Results #######################");
        Serial.println("COLOR\t|\tcorrect\t|\twrong\t|\tmax(ms)\t|\tmin(ms)\t|\tmean(ms)\t|\tdevstd(n/a)");
        Serial.print("RED:\t|\t"); R_stats.printStats();
        Serial.print("GREEN:\t|\t"); G_stats.printStats();
        Serial.print("BLUE:\t|\t"); B_stats.printStats();
        Serial.println("######################################################");
        Serial.println();

        tone(3, 4000, 1500);
        for (int i = 0; i < 10; i++) {
          pRGB->blinkBlue(100, 50);
        }
        pRGB->blinkRed(500);
        mMeasFinished = true;
      }
    }

    bool measFinished(){
      return mMeasFinished;
    }

    void disable(){
      mEnabled = false;
    }
    void enable(){
      mEnabled = true;
    }

    bool enabled(){
      return mEnabled;
    }
    
    void deinit(){
      mMeasFinished = true;
    }

    String getStatsJSON(){
      String data = "";
      data += "{ \"EQUID\": \"" + String(mEQUID);
      data += ", \"CRT_R_STATS\":" + R_stats.getStats_asJSONLine();
      data += ", \"CRT_G_STATS\":" + G_stats.getStats_asJSONLine();
      data += ", \"CRT_B_STATS\":" + B_stats.getStats_asJSONLine();
      
      if (pBattery != NULL)
        data += ", \"CRT_BATTERY\":" + String(pBattery->getBatteryPercent());
      data += "}";

      return data;
    }
};
*/
