/*
 *  Arduino library for Moodmetric Ring
 *  Version _2.0
 *  @author Calin Bira (calin.bira@upb.ro) 
 */

#include <ArduinoBLE.h>
#include <limits.h>
#include "issUtils.hpp"

class issMoodmetricRing {

    BLEDevice peripheral;
    int mBatteryPercent = 0;
    int mMM = 0;
    float mAccXYZg[3] = {0, 0, 0};
    bool mMMElevated = false;
    bool mMMLow = false;
    bool mMMnotReady = true;
    uint16_t mSkinResReg = -1;
    int mEQUID;

    issSmartRGB* pRGB;
    issBattery *pBattery = NULL;
    
    issAvg<uint32_t> skinres;
    issAvg<uint8_t> mmindex;
          
    int BlinkMs = 5;

  public:
    issMoodmetricRing(issSmartRGB *pled, issBattery *pbat, int equid) {
      mEQUID = equid;
      pRGB = pled;
      pBattery = pbat;
    }

    bool init() {
      // begin initialization.
      for (int i = 0; i < 10; i++) {
        if (BLE.begin()) return true;
        delay(500);
      }
      Serial.println("[ERROR] BLE init failed");
      return false;
    }

    bool scanForRing(String name) {
      if (!init()) return false;
      BLE.scanForName(name);
      for (int i = 0; i < 10; i++) {
        peripheral = BLE.available();
        if (peripheral) {
          BLE.stopScan();
          return true;
        }

        delay(500);
      }

      BLE.stopScan();
      return false;
    }

    bool update(String RingName, String RingNameAlt) {
      Serial.println("\n\n<MM>[INFO] Scanning for ring " + RingName + " and " + RingNameAlt);
      if (!scanForRing(RingName) && !scanForRing(RingNameAlt)) {
        Serial.println("<MM>NOT discovered");
        pRGB->blinkRed(BlinkMs, BlinkMs);
        BLE.end();
        return false;
      }
      Serial.println("<MM>discovered");
      Serial.print("<MM>[INFO] Connecting to ring... ");
      if (!peripheral.connect()) {
        Serial.println("could not connect to ring");
        pRGB->blinkRed(BlinkMs, BlinkMs);
        BLE.end();
        return false;
      }
      Serial.println("connected to ring");

      peripheral.discoverAttributes();
      BLEService measurementService = peripheral.service("dd499b70-e4cd-4988-a923-a7aab7283f8e");

      // StreamingCharacteristic: a0956420-9bd2-11e4-bd06-0800200c9a66
      // NotificationCharacteristic: c48650d0-a2d8-11e4-bcd8-0800200c9a66
      // RawDataCharacteristic: f1b41cde-dbf5-4acf-8679-ecb8b4dca6ff
      BLECharacteristic RawDataCharacteristic = measurementService.characteristic("f1b41cde-dbf5-4acf-8679-ecb8b4dca6ff");
      long startms = millis();

      skinres.clear();
      while (millis() - startms < MM_PARAM_SWEEP_TIME_SECONDS*1000) {
        if (RawDataCharacteristic.canRead()) {
          delay(300);
          if (true == RawDataCharacteristic.read()) {
            if (RawDataCharacteristic.valueLength() >= 2) {
              mSkinResReg = RawDataCharacteristic.value()[0]*256 + RawDataCharacteristic.value()[1];
              skinres.pushValue(getSkinResistanceKOhm());
              Serial.println("<MM>Skin resistance (16-bit value) is: " + (String)mSkinResReg);
              Serial.println("<MM>Skin resistance (KOhm) is: " + (String)getSkinResistanceKOhm()); 
            }
            else {
              Serial.println("<MM>bad size for RawDataCharacteristic data");
              break;
            }
          }
        } 
        else {
          Serial.println("<MM>unsupported RawDataCharacteristic");
          break;
        }
      }

      //
      BLECharacteristic StreamingCharacteristic = measurementService.characteristic("a0956420-9bd2-11e4-bd06-0800200c9a66");

      mmindex.clear();
      startms = millis();
      while (millis() - startms < MM_PARAM_SWEEP_TIME_SECONDS * 1000)
        if (StreamingCharacteristic.canRead()) {
          delay(300);
          if (true == StreamingCharacteristic.read()) {
            if (StreamingCharacteristic.valueLength() >= 7) {
              unsigned char status = StreamingCharacteristic.value()[0];
              mMM = StreamingCharacteristic.value()[1];
  
              //mInstant = StreamingCharacteristic.value()[2]*256 + StreamingCharacteristic.value()[3];
              
              mAccXYZg[0] = (float)StreamingCharacteristic.value()[4] / 128 * 2 - 2;
              mAccXYZg[1] = (float)StreamingCharacteristic.value()[5] / 128 * 2 - 2;
              mAccXYZg[2] = (float)StreamingCharacteristic.value()[6] / 128 * 2 - 2;
  
              String statusbits = "<MM>";
              
              if (status & 1) statusbits += ("| MM not ready"); else statusbits += ("| MM ready");
              if (status & 2) statusbits += ("| Ring on finger");
              if (status & 4) statusbits += ("| Battery is out");
              if (status & 8) statusbits += ("| Any reaction");
              if (status & 16) statusbits += ("| Strong reaction");
              if (status & 32) statusbits += ("| MM elevated for 5 minutes");
              if (status & 64) statusbits += ("| MM low for 5 minutes");
              if (statusbits.length() > 1) {
                statusbits += "|";
                Serial.println(statusbits);
              }
    
              mMMnotReady = (status & 1);
              mMMElevated = status & 32;
              mMMLow = status & 64;
    
              Serial.println("<MM>MM number is " + String(mMM) + ", ACC is " + String(mAccXYZg[0]) + "g, " + String(mAccXYZg[1]) + " g, " + String(mAccXYZg[2]) + " g");

              if (!mMMnotReady)
                mmindex.pushValue(mMM);
            }
            else {
              Serial.println("<MM>[WARNING] Unable to read data characteristic");
              peripheral.disconnect();
              //        BLE.end();
              return 0;
            }
          }
        }
        else {
          Serial.println("<MM>[WARNING] Unable to read data characteristic");
          peripheral.disconnect();
          //        BLE.end();
          return 0;
        }
      
      BLEService batteryService = peripheral.service("180f");
      BLECharacteristic batteryCharacteristic = batteryService.characteristic("2a19");

      if (batteryCharacteristic.canRead()) {
        batteryCharacteristic.read();
        mBatteryPercent = batteryCharacteristic.value()[0];
        Serial.println("<MM>Battery at " + String(mBatteryPercent) + " %");
      }
      else
        Serial.println("<MM>[WARNING] Unable to read battery characteristic");

      peripheral.disconnect();
      //      BLE.end();
      if (mMMnotReady) pRGB->blinkBlue(BlinkMs, BlinkMs);
      Serial.println(String ("<MM>") + String("mMMnotReady: ") + String((mMMnotReady ? "true" : "false")));
      return !mMMnotReady;
    }

    int getBatteryLevel() {
      return mBatteryPercent;
    }

    bool getMMElevated() {
      return mMMElevated;
    }

    bool getMMLow() {
      return mMMLow;
    }

    int getMM() {
      return mMM;
    }

    uint32_t getSkinResistanceKOhm() {
      return mSkinResReg * (16000000UL / 65536UL) / 1000;
    }

    void printAcceleration() {
      Serial.println("x-" + String(mAccXYZg[0]) + "\t y-" + String(mAccXYZg[1]) + "\t z-" + String(mAccXYZg[2]));
    }

    void exploreService(BLEService service) {
      // print the UUID of the service
      Serial.print("<MM>Service ");
      Serial.println(service.uuid());

      // loop the characteristics of the service and explore each
      for (int i = 0; i < service.characteristicCount(); i++) {
        BLECharacteristic characteristic = service.characteristic(i);

        exploreCharacteristic(characteristic);
      }
    }

    void exploreCharacteristic(BLECharacteristic characteristic) {
      // print the UUID and properties of the characteristic
      Serial.print("\tCharacteristic ");
      Serial.print(characteristic.uuid());
      Serial.print(", properties 0x");
      Serial.print(characteristic.properties(), HEX);

      // check if the characteristic is readable
      if (characteristic.canRead()) {
        // read the characteristic value
        characteristic.read();

        if (characteristic.valueLength() > 0) {
          // print out the value of the characteristic
          Serial.print(", value 0x");
          printData(characteristic.value(), characteristic.valueLength());
        }
      }
      Serial.println();

      // loop the descriptors of the characteristic and explore each
      for (int i = 0; i < characteristic.descriptorCount(); i++) {
        BLEDescriptor descriptor = characteristic.descriptor(i);

        exploreDescriptor(descriptor);
      }
    }

    void exploreDescriptor(BLEDescriptor descriptor) {
      // print the UUID of the descriptor
      Serial.print("\t\tDescriptor ");
      Serial.print(descriptor.uuid());

      // read the descriptor value
      descriptor.read();

      // print out the value of the descriptor
      Serial.print(", value 0x");
      printData(descriptor.value(), descriptor.valueLength());

      Serial.println();
    }

    void printData(const unsigned char data[], int length) {
      for (int i = 0; i < length; i++) {
        unsigned char b = data[i];

        if (b < 16) {
          Serial.print("0");
        }

        Serial.print(b, HEX);
      }
    }

    String getStatsJSON(int duration_s) {
      String data = "";
      data += "{";
      data += "\"NETWORK_DELAY\": " + String(duration_s) + ",";
      data += "\"EQUID\": \"" + String(mEQUID) + "\", ";
      data += "\"MM_INDEX\": " + String(mmindex.getAvg()) + ", ";
      data += "\"MM_RESISTANCE\": " + String(skinres.getAvg()) + ", ";
      data += "\"MOOD_LOW_FLAG\": " + String(mMMLow) + ", ";
      data += "\"MOOD_HIGH_FLAG\": " + String(mMMElevated) + ", ";
      if (pBattery != NULL)
        data += "\"CRT_BATTERY\":" + String(pBattery->getBatteryPercent()) + ", ";
      data += "\"MM_BATTERY\":" + String(mBatteryPercent);
      data += "}";

      return data;
    }
};
