/*
 * Arduino software for ISS wellBeing smart system
 * @author Costin Vasile [CV] (costin.vasile1003@upb.ro)
 * @author Calin Bira [CLN] (calin.bira@upb.ro)
 * 
 * TODO:
 *    - add LowPower support
 *    - update MMR data to cloud only when change (now implemented to send after each measurement)
 *    - software documentation: install Arduino SAMD (board wifi1010), WifiNina (lib), ArduinoBLE (lib), SAMD_TimerInterrupt, WDTZero (Internet)
 *    
 * v1.0 26.05.2021 - CV  initial version for Choice Reaction Time (CRT) measurement / Moodmetric Ring measurement / Data sending to cloud
 * v1.1 19.10.2021 - CLN updated with issBattery
 * v1.2 02.11.2021 - CLN implemented as ISR-driven FSM (so that issWellBeing to be able to read ring and do CRT in the "same" time)
 * v1.3 03.11.2021 - CLN docu update, bugfix: fake led lit-up avoided by waiting buttons to recover from bouncing
 * v1.4 04.11.2021 - CLN updated with NETWORK_DELAY field in json; increased issAvg sum to uin32_t;
 * v1.5 10.08.2022 - CLN updated with <CRT> and <MM> tags
 * v1.6 10.08.2022 - CLN updated with first correct and first mistakenly pressed button in stats
 * v1.7 11.08.2022 - CLN bugfix CRT wrong-right-wrong pressing scenario
 * v1.8 17.08.2022 - CLN bugfix CRT wrong++ incremented once per entry (not continuously)
 * v1.9 07.09.2022 - CLN updated comments
 */
String sw_version = "v1.9";

// Management interface is at:
// http://saveaal.eu:4200/admincentre

#include "issWiFi.hpp"
#include <WDTZero.h>

#define MOODMETRIC3
//#define MOODMETRIC4
//#define MOODMETRIC5
//#define MOODMETRIC6
//#define MOODMETRIC7
//#define MOODMETRIC9

#include "issSystemConfig.hpp"
#include "issBattery.hpp"
#include "issSmartRGB.hpp"
#include "issSpeaker.hpp"
#include "issMM.hpp"
#include "issCRT2.hpp" 

#define RETRY_DATA_SENDING_MM_TIMES 2

long start = millis() - MM_POLLING_TIME_EVERY_X_SECONDS * 1000;

WDTZero wdt;
issBattery batt;
issSmartRGB rgb(CRT_LED_R, CRT_LED_G, CRT_LED_B, 100);

issWifi wifi(ssid, pass);
issWebClient client(&wifi, server);
issMoodmetricRing mmr(&rgb, &batt, mmr_equid);

issSpeaker speaker(SPEAKER_PIN);
issCRT2 crt(&rgb, &batt, &speaker, CRT_BUTTON_LEFT, CRT_BUTTON_RIGHT, CRT_RANDOM_PIN, crt_equid);

void setup() {
  Serial.begin(115200);
  Serial.println("<MM><CRT> Software version " + sw_version);
  Serial.println("<MM> Using moodmetric ring with index: " + String(MOODMETRIC_INDEX));

  rgb.showAllWork();
  
  setupTimer0();
  wdt.setup(WDT_SOFTCYCLE2M);
}

void loopInterrupt() {
  crt.interruptLoop();
}

void loop() {
  wdt.clear();
  
  //NB: crt works on interrupts  
  if (crt.measFinished()) {
    bool result = false;
    while (result == false) {
      String body = crt.getStatsJSON((millis() - crt.getStartTimestamp())/1000);
      Serial.println(body);
      result = client.do_POST(body.c_str(), body.length(), path, crt_apiKey, crt_port);
    }
    ; // blocking wait for data to be sent
    
    // confirmation that data was sent
    speaker.doSoundDataWasSent();
    
    crt.reset(); // prepare for next CRT experiment
  }

  //if (!wifi.isConnected()) wifi.reconnect();

  if ((millis()-start) > MM_POLLING_TIME_EVERY_X_SECONDS * 1000) {
   wifi.disconnect();
   long startMMmilliseconds = millis();
   if (mmr.update(MMRingName, MMRingNameAlt)) {
    Serial.println();
    Serial.println("############### New MMR data ###############");
    Serial.println("MM: " + String(mmr.getMM()));
    Serial.println("Battery: " + String(mmr.getBatteryLevel()) + "%");
    Serial.print("Acceleration:\t"); mmr.printAcceleration();
    Serial.println("Elevated for more than 5 minutes:\t" + String(mmr.getMMElevated()));
    Serial.println("Inactive for more than 5 minutes:\t" + String(mmr.getMMLow()));
    Serial.println();

    //if (true == client.do_POST(body.c_str(), body.length(), path, mmr_apiKey, mmr_port))
      //rgb.blinkGreen();

    uint8_t retry = RETRY_DATA_SENDING_MM_TIMES;
    while (retry-- > 0) {
      String body = mmr.getStatsJSON((millis()-startMMmilliseconds)/1000);
      Serial.println(body);
      if (true == client.do_POST(body.c_str(), body.length(), path, mmr_apiKey, mmr_port))
        break;
    }
      
    Serial.println("###########################################");
    start = millis();
  }
  }
}

#if !( defined(ARDUINO_SAMD_ZERO) || defined(ARDUINO_SAMD_MKR1000) || defined(ARDUINO_SAMD_MKRWIFI1010) \
    || defined(ARDUINO_SAMD_NANO_33_IOT) || defined(ARDUINO_SAMD_MKRFox1200) || defined(ARDUINO_SAMD_MKRWAN1300) || defined(ARDUINO_SAMD_MKRWAN1310) \
    || defined(ARDUINO_SAMD_MKRGSM1400) || defined(ARDUINO_SAMD_MKRNB1500) || defined(ARDUINO_SAMD_MKRVIDOR4000) || defined(__SAMD21G18A__) \
    || defined(ARDUINO_SAMD_CIRCUITPLAYGROUND_EXPRESS) || defined(__SAMD21E18A__) || defined(__SAMD51__) || defined(__SAMD51J20A__) || defined(__SAMD51J19A__) \
    || defined(__SAMD51G19A__) || defined(__SAMD51P19A__) || defined(__SAMD21G18A__) )
  #error This code is designed to run on SAMD21/SAMD51 platform! Please check your Tools->Board setting.
#endif
#define TIMER_INTERRUPT_DEBUG         0
#define _TIMERINTERRUPT_LOGLEVEL_     0
#include <SAMDTimerInterrupt.h>
#define TIMER0_INTERVAL_MS 1

SAMDTimer ITimer0(TIMER_TC3);

void setupTimer0() {
  if (ITimer0.attachInterruptInterval(TIMER0_INTERVAL_MS * 1000, TimerHandler0))
  {
    long preMillisTimer0 = millis();
    Serial.print(F("Starting  ITimer0 OK, millis() = ")); Serial.println(preMillisTimer0);
  }
}

void TimerHandler0()
{
  loopInterrupt();
}
