/*
 *  Arduino library used to handle a WiFi client used for ISS WellBeing system
 *  Version _1.0
 *  
 *  @author Costin Vasile (costin.vasile1003@upb.ro)
 */

#include <SPI.h>
#include <WiFiNINA.h>

class issWifi {
  private:
    char* mSSID;
    char* mPASS;
    bool mConnected;
    
  public:
    issWifi(char* SSID, char* PASS) {
        mSSID = SSID;
        mPASS = PASS;
        mConnected = false;
    }

    bool connect() {
      uint8_t tries = 0;
      int status = WL_IDLE_STATUS;
      while(status != WL_CONNECTED) {
        if(tries > 5) {
          Serial.print("<WIFI>[WARNING] Unable to connect to SSID: ");
          Serial.print(mSSID);
          Serial.println(" | WiFi functionality will not be available!");
          return false;
        }
        Serial.print("<WIFI>[INFO] Attempting to connect to SSID: ");
        Serial.println(mSSID);
        status = WiFi.begin(mSSID, mPASS);
        tries++;
        delay(2000 * tries);
      }
      Serial.println("<WIFI>[INFO] Connected to WiFi!");
      return true;
    }

    bool isConnected() {
      return WiFi.status() == WL_CONNECTED;
    }

    void disconnect() {
      WiFi.end();
    }

    bool reconnect() {
      disconnect();      
      return connect();
    }
    
    void printWifiStatus() {
      Serial.print("<WIFI>[INFO] SSID: ");
      Serial.println(WiFi.SSID());
      
      IPAddress ip = WiFi.localIP();
      Serial.print("<WIFI>[INFO] IP Address: ");
      Serial.println(ip);
      
      long rssi = WiFi.RSSI();
      Serial.print("<WIFI>[INFO] Signal strength (RSSI):");
      Serial.print(rssi);
      Serial.println(" dBm");
    }
};

class issWebClient{
 private:
  char* mServer;
  WiFiClient mClient;
  issWifi *pWifi;

 public:
  issWebClient(issWifi *wifi, char* server){
    mServer = server;
    pWifi = wifi;
  }
  
  bool do_POST(const char* body, uint8_t body_size, char* path, char* api_key, uint8_t port){
    if (!pWifi->isConnected()) pWifi->reconnect();
    if(!pWifi->isConnected()){
      Serial.println("<WIFI>[WARNING] WiFi not connected. Unable to send data to cloud");
      pWifi->disconnect();
      return false;
    }
    
    Serial.print("<WIFI>[INFO] Connecting to server... ");
    if(!mClient.connect(mServer, port)){
      Serial.println("\n<WIFI>[WARNING] Connection failed! Data not sent");
      pWifi->disconnect();
      return false;
    }
    Serial.println ("<WIFI>connection done");

    //Serial.println();
    mClient.println("POST " + String(path) + " HTTP/1.0"); //Serial.println("POST " + String(path) + " HTTP/1.0");
    mClient.println("x-api-key: " + String(api_key)); //Serial.println("x-api-key: " + String(api_key));
    mClient.println("Content-Type: application/json"); //Serial.println("Content-Type: application/json");
    mClient.println("Content-Length: " + String(body_size)); //Serial.println("Content-Length: " + String(body_size));
    mClient.println("Host: www.example.com"); //Serial.println("Host: www.example.com");
    mClient.println(); //Serial.println();
    mClient.println(String(body)); //Serial.println(String(body));
    mClient.println(); //Serial.println();
    //Serial.println();

    int waitLoops = 50;
    while (!mClient.available() && (waitLoops > 0)) {
      waitLoops--;
      delay(100);
    }
    
    if (waitLoops == 0) {
      Serial.println("<WIFI>[WARNING] No answer from server. Data might not be sent!");
    }

    char* answer = (char*) malloc(12*sizeof(char));
    for(int i = 0; i < 12; i++){
      if(!mClient.available()){
        pWifi->disconnect();
        return false;
      }
      answer[i] = mClient.read();
    }

    if(!strcmp(answer, "HTTP/1.1 200")){
      Serial.println("<WIFI>[WARNING] The server answered, but with no success response. Data might not be sent!");
      return false;
    }
    
    Serial.println("<WIFI>[INFO] Data sent to cloud");
    pWifi->disconnect();
    return true;
  }
};
