/**
 * A class implementing a pulled-up button
 * Pin will be HIGH when button is not pressed, and LOW when it is pressed
 * Can additionally detect continuous pressing for milliseconds
 *
 * @author Calin Bira (calin.bira@upb.ro)
 */
 
#pragma once

class issButton {
  
  int mPin;  //!< pin number where the button is connected

  int mConsecutiveHighTimes; /**< milliseconds while button is not pressed (resets mConsecutiveLowTimes) */
  int mConsecutiveLowTimes;  /**<  milliseconds while button is pressed (resets mConsecutiveHighTimes) */
  long mTimestampLastEvent;  /**<  internal timestamp in milliseconds */
  const int INTERVAL_MS = 3; /**<  in milliseconds, used to lower cpu usage and for debouncing */
  
  public:
    /**
     * Constructor
     * @param pin pin number
     */
    issButton(int pin) {
      mPin = pin;
      pinMode(mPin, INPUT_PULLUP);
      mConsecutiveHighTimes = 0;
      mConsecutiveLowTimes = 0;
      mTimestampLastEvent = millis();
    }

    /**
     * Method to be called from a time-based interrupt (can also be called from non-isr code)
     * Updates internal data regarding consecutive pressing
     */
    void issButtonloop() {
      issTimerTasks::getInstance()->loop();
      if (millis() - mTimestampLastEvent >= INTERVAL_MS) {
        int value = digitalRead(mPin);    
        if (value == HIGH) {mConsecutiveHighTimes++; mConsecutiveLowTimes = 0; mTimestampLastEvent = millis();}
        if (value == LOW) {mConsecutiveLowTimes++; mConsecutiveHighTimes = 0; mTimestampLastEvent = millis();}
      }
    }

  /**
   * Checks if button is pressed or not
   * @return true if pin is low, false otherwise
   */
    bool isButtonPressed() {
      return digitalRead(mPin) == LOW;
    }

    static inline bool isButtonPressed(int pin) {
      return digitalRead(pin) == LOW;
    }

  /**
   * Checks number of continuous HIGH in milliseconds
   * @return milliseconds multiple of INTERVAL_MS
   */
    int getMillisContinuousHigh() {
        return INTERVAL_MS * mConsecutiveHighTimes;
    }

  /**
   * Checks number of continuous LOW in milliseconds
   * @return milliseconds multiple of INTERVAL_MS
   */
    int getMillisContinuousLow() {
        return INTERVAL_MS * mConsecutiveLowTimes;
    }    

  /**
   * Checks number of continuous press in milliseconds
   * @return milliseconds multiple of INTERVAL_MS
   */
    int getMillisContinuousPressed() {
        return getMillisContinuousLow();
    }
};
