/*
 *  This file contains issCRTStats class used to handle Choice Reaction Time (CRT) data packets
 *  Version _1.1
 *  
 *  @author Costin Vasile (costin.vasile1003@upb.ro)
 *  @author Calin Bira (calin.bira@upb.ro)
 */
#pragma once

class issCRTStats {
    volatile int mWrong;
    volatile int mCorrect;
    volatile int mMeasCounter;
    volatile int mSamplesCnt;
    
    unsigned mMax;
    unsigned mMin;
    unsigned mMean;
    int mDevStd;

    #define MAX_SAMPLES 30
    int mCorrectSamples[MAX_SAMPLES]; // timestamp of correct pressing
    int mWrongSamples[MAX_SAMPLES]; // timestamp of first press, regardless if it is correct or wrong


  public:
    issCRTStats() {
      resetStats();
    }

    void resetStats() {
      mWrong = 0;
      mCorrect = 0;
      mMax = 0;
      mMin = -1;
      mMean = 0;
      mDevStd = 0;
      mMeasCounter = 0;
      mSamplesCnt = 0;
      for (int i = 0; i < MAX_SAMPLES; i++) {
        mCorrectSamples[i] = 0;
        mWrongSamples[i] = 0;
      }
    }

    void updateStdDev(){
      float stdDev = 0;
      
      for(int i = 0; i<mSamplesCnt; i++){
        stdDev += sq(mCorrectSamples[i] - mMean);
        if (mWrongSamples[i] != 0)
          stdDev += sq(mWrongSamples[i] - mMean);
      }

      stdDev = sqrt( stdDev / (mSamplesCnt - 1) );
      mDevStd = stdDev;
    }

    unsigned getlastWrongTime() { return mWrongSamples[mSamplesCnt];}
    unsigned getlastCorrectTime() { return mCorrectSamples[mSamplesCnt];}
    void advance() { if (mSamplesCnt < MAX_SAMPLES) mSamplesCnt++; }
    
    void updateStats(bool iscorrect, unsigned responseTime) {
      mMeasCounter++;
      if (iscorrect) {
        mCorrect++;
        mCorrectSamples[mSamplesCnt] = responseTime;
      }
      else {
        if (mWrongSamples[mSamplesCnt]==0) mWrong++; // increment only once per record
        mWrongSamples[mSamplesCnt] = responseTime;
      }
      
      if (responseTime < mMin) mMin = responseTime;
      if (responseTime > mMax) mMax = responseTime;

      //if (mMeasCounter == 0) mMean = 0;
      //else 
      if (mMeasCounter == 1) mMean = responseTime;
      else mMean = ((mMean * (mMeasCounter - 1)) + responseTime) / mMeasCounter;

      updateStdDev();
    }

    bool hasData() {
      return mMeasCounter > 0;
    }

    void printlnStats(){
      Serial.print(mCorrect);
      Serial.print("\t|\t");
      Serial.print(mWrong);
      Serial.print("\t|\t");
      Serial.print(mMax);
      Serial.print("\t|\t");
      Serial.print(mMin);
      Serial.print("\t|\t");
      Serial.print(mMean);
      Serial.print("\t|\t");
      Serial.println(mDevStd);
    }

    inline String getStats_asJSONLine(){
      String JSONLine = "[" + String(mWrong) + "," + String(mCorrect) + "," + String(mMin) + "," + String(mMax) + "," + String(mMean) + "," + String(mDevStd) + "]";

      return JSONLine;
    }
};
