/*
 * Arduino software for ISS wellBeing smart system
 * @author Costin Vasile [CV] (costin.vasile1003@upb.ro)
 * @author Calin Bira [CLN] (calin.bira@upb.ro)
 */

/*
  1) Device starts by looping through RED, GREEN, BLUE leds.
  2) Device rests in IDLE state
  3) From IDLE (solid RED) state it goes to CRT_IDLE (solid GREEN) on 2-button press
  4) From CRT_IDLE (solid GREEN) it goes to MM_IDLE (solid BLUE) on 2-button press
  5) From MM_IDLE (solid BLUE) it goes to RNG_IDLE (solid VIOLET) on 2-button press
  
  Therefore: on both button pressed, device cycles through IDLE -> CRT_IDLE -> MM_IDLE ->IDLE ...
  6) From CRT_IDLE: if LEFT button is pressed it goes to CRT_LOOP and CRT experiment starts
  7) From MM_IDLE: if LEFT button is pressed it goes to MM_LOOP and MM poling of data starts
  8) From RNG_IDLE: if LEFT button is pressed it goes to RNG_LOOP and rng data generation starts 
  After each CRT_LOOP or MM_LOOP, data is transmitted wia wifi
  When led is solid WHITE, wifi is trying to send data
 */

/*
 * TODO:
 *    - add LowPower support
 *    - update MMR data to cloud only when change (now implemented to send after each measurement)
 *    - software documentation: install Arduino SAMD (board wifi1010), WifiNina (lib), ArduinoBLE (lib), SAMD_TimerInterrupt, WDTZero (Internet)
 *    
 * v1.0 26.05.2021 - CV  initial version for Choice Reaction Time (CRT) measurement / Moodmetric Ring measurement / Data sending to cloud
 * v1.1 19.10.2021 - CLN updated with issBattery
 * v1.2 02.11.2021 - CLN implemented as ISR-driven FSM (so that issWellBeing to be able to read ring and do CRT in the "same" time)
 * v1.3 03.11.2021 - CLN docu update, bugfix: fake led lit-up avoided by waiting buttons to recover from bouncing
 * v1.4 04.11.2021 - CLN updated with NETWORK_DELAY field in json; increased issAvg sum to uin32_t;
 * v1.5 10.08.2022 - CLN updated with <CRT> and <MM> tags
 * v1.6 10.08.2022 - CLN updated with first correct and first mistakenly pressed button in stats
 * v1.7 11.08.2022 - CLN bugfix CRT wrong-right-wrong pressing scenario
 * v1.8 17.08.2022 - CLN bugfix CRT wrong++ incremented once per entry (not continuously)
 * v1.9 07.09.2022 - CLN updated comments
 * v2.0 12.02.2023 - CLN cooperative multitasking + software polled timers to avoid interrupts
 * v2.1 09.12.2023 - CLN added RNG number generation for NIST tests
 */
 
String sw_version = "v2.1";

// Management interface is at:
// http://saveaal.eu:4200/admincentre

#include "issWiFi.hpp"
//#define MOODMETRIC3
//#define MOODMETRIC4
#define MOODMETRIC5
//#define MOODMETRIC6
//#define MOODMETRIC7
//#define MOODMETRIC9

#include "issSystemConfig.hpp"
#include "issBattery.hpp"
#include "issSpeaker.hpp"

#include "issTimerTasks.hpp"
#include "issSmartLED.hpp"
#include "issSmartRGB.hpp"
#include "issMM.hpp"
#include "issCRT3.hpp" 
#include "issTest.hpp"

typedef enum WellBeingState {
  WB_IDLE = 0, //
  
  WB_CRT_IDLE = 1,
  WB_CRT_LOOP = 2,
  
  WB_MM_IDLE = 3,  
  WB_MM_LOOP = 4,

  WB_RNG_IDLE = 5,
  WB_RNG_LOOP = 6
} WellBeingState_t;

#define RETRY_DATA_SENDING_MM_TIMES 2
#define RETRY_DATA_SENDING_CRT_TIMES 2
long start = millis() - MM_POLLING_TIME_EVERY_X_SECONDS * 1000;

issBattery batt;
issSmartRGB rgb(CRT_LED_R, CRT_LED_G, CRT_LED_B, 100);

issWifi wifi(ssid, pass);
issWebClient client(&wifi, server);
issMoodmetricRing mmr(&rgb, &batt, mmr_equid);

issSpeaker speaker(SPEAKER_PIN);
issCRT3 crt(&rgb, &batt, &speaker, CRT_BUTTON_LEFT, CRT_BUTTON_RIGHT, CRT_RANDOM_PIN, crt_equid);

issTimerTask* tprint;
issTimerTask* twbstate;

WellBeingState_t wbs;

void setup() {
  Serial.begin(115200);
  Serial.println("<MM><CRT> Software version " + sw_version);
  Serial.println("<MM> Using moodmetric ring with index: " + String(MOODMETRIC_INDEX));

  rgb.showAllWork();
  tprint = new issTimerTask(5000);
  tprint->restartTimerTask();

  twbstate = new issTimerTask(2000);
  twbstate->restartTimerTask();
  //setupTimerTest();

  wbs = WB_IDLE;
}

void loop() {
  //NB: crt works on interrupts  
  issTimerTasks::getInstance()->loop();
  
  switch (wbs) {
    case WB_IDLE: {
      rgb.switchAllOff(); rgb.switchRedOn();
      twbstate->mwait();
      if (crt.areButtonsPressed()) {
        wbs = WB_CRT_IDLE;
        Serial.println("<SYS> CRT mode?");
      }
      break;
    }
    case WB_CRT_IDLE: {
      rgb.switchAllOff();
      rgb.switchGreenOn();
      twbstate->mwait();
      wbs = WB_CRT_LOOP;
      break;
    }
    case WB_CRT_LOOP: {
      if ((crt.areButtonsPressed()) && (crt.measFinished())) {wbs = WB_IDLE; Serial.println("<SYS> IDLE mode");}
      else if (crt.areButtonsPressed()) {wbs = WB_MM_IDLE; Serial.println("<SYS> MM mode");}
      break;
    }
    case WB_MM_IDLE: {
      rgb.switchAllOff(); 
      rgb.switchBlueOn();
      twbstate->mwait();
      wbs = WB_MM_LOOP;
      break;
    }
    case WB_MM_LOOP: {
      crt.issButtonloop();
      // goes automatically to RNG after MM experiment
      if (crt.areButtonsPressed()) {wbs = WB_RNG_IDLE; Serial.println("<SYS> RNG mode");}
      break;
    }
    case WB_RNG_IDLE: {
      rgb.switchAllOff(); 
      rgb.switchBlueOn();
      rgb.switchRedOn();
      twbstate->mwait();

      if (crt.areButtonsPressed()) {
        wbs = WB_IDLE;
        Serial.println("<SYS> IDLE mode");
      }
      else
        wbs = WB_RNG_LOOP;
      break;
    }

    default:
      break;
  }
 
  if (wbs == WB_IDLE) {
    //if (tprint->isOverflown()) {
    //  issTimerTasks::getInstance()->printAll();
    //  tprint->clearOverflow();
    //}
  } 
  else if (wbs == WB_CRT_LOOP) {
    crt.issCRTloop();  
    if (crt.measFinished()) {
      Serial.println("<SYS> shutting down ble, sending via wifi-crt");
      bool result = false;

      uint8_t retry = RETRY_DATA_SENDING_CRT_TIMES;
      while (retry-- > 0) {
        rgb.switchAllOff();
        String body = crt.getStatsJSON((millis() - crt.getStartTimestamp())/1000);
        Serial.println(body);
        result = client.do_POST(body.c_str(), body.length(), path, crt_apiKey, crt_port);
        rgb.switchAllOn();
        if (crt.areButtonsPressed()) {wbs = WB_IDLE; break;}
        if (result == true) break;
      }
      ; // blocking wait for data to be sent
      
      // confirmation that data was sent
      speaker.doSoundDataWasSent();    
      crt.reset(); // prepare for next CRT experiment
      wbs = WB_MM_IDLE;
    }
  }
  //if (!wifi.isConnected()) wifi.reconnect();
  else if (wbs == WB_MM_LOOP) {
    if (crt.isRightButtonPressed(2000))
    if ((millis()-start) > MM_POLLING_TIME_EVERY_X_SECONDS * 1000) {
      Serial.println("<SYS> shutting down wifi, checking BLE");
      wifi.disconnect();
      long startMMmilliseconds = millis();
      if (mmr.update(MMRingName, MMRingNameAlt)) {
        Serial.println();
        Serial.println("<SYS>############### New MMR data ###############");
        Serial.println("<SYS>MM: " + String(mmr.getMM()));
        Serial.println("<SYS>Battery: " + String(mmr.getBatteryLevel()) + "%");
        Serial.print  ("<SYS>Acceleration:\t"); mmr.printAcceleration();
        Serial.println("<SYS>Elevated for more than 5 minutes:\t" + String(mmr.getMMElevated()));
        Serial.println("<SYS>Inactive for more than 5 minutes:\t" + String(mmr.getMMLow()));
        Serial.println();
    
        //if (true == client.do_POST(body.c_str(), body.length(), path, mmr_apiKey, mmr_port))
          //rgb.blinkGreen();
        Serial.println("<SYS> shutting down ble, sending via wifi-mm");
        uint8_t retry = RETRY_DATA_SENDING_MM_TIMES;

        rgb.switchAllOff(); rgb.switchAllOn();
        while (retry-- > 0) {
          String body = mmr.getStatsJSON((millis()-startMMmilliseconds)/1000);          
          Serial.println(body);
          if (true == client.do_POST(body.c_str(), body.length(), path, mmr_apiKey, mmr_port))
            break;
          if (crt.areButtonsPressed()) {wbs = WB_IDLE; break;}
        }
        
        Serial.println("<SYS>###########################################");
        start = millis();
      }
      wbs = WB_IDLE;
    }
  }
  if (wbs == WB_RNG_LOOP) {
    issRandom* pRandom = new issRandom(CRT_RANDOM_PIN);
    for (int i = 0; i < 2000*1000; i++) {
      int randval = pRandom->getRandomNumber(CRT_LED_START, CRT_LED_STOP+1);
      Serial.println(randval);
    }
    delete(pRandom);
    wbs = WB_IDLE;      
  }
}
