#pragma once

#include <Wire.h>
#include <MPU6050.h>

class issWxAccelerometer {
  /*
    MPU6050 Triple Axis Gyroscope & Accelerometer. Motion detection.
    Read more: http://www.jarzebski.pl/arduino/czujniki-i-sensory/3-osiowy-zyroskop-i-akcelerometr-mpu6050.html
    GIT: https://github.com/jarzebski/Arduino-MPU6050
    Web: http://www.jarzebski.pl
    (c) 2014 by Korneliusz Jarzebski
*/
#define PIN_LED1 6
#define PIN_LED2 13
#define PIN_SPEAKER 9

//#define ENABLE_SOUND
//#define ENABLE_LEDS
//#define ENABLE_VERBOSE_ACC

MPU6050 mpu;
Vector lastrawAccel;
  
public:

/*
  * LP_WAKE_CTRL | Wake-up Frequency
  * -------------+------------------
  * 0            | 1.25 Hz
  * 1            | 2.5 Hz
  * 2            | 5 Hz
  * 3            | 10 H

*/

#define  LP_WAKE_CTRL_1_25        0x00
#define  LP_WAKE_CTRL_2_5         0x01
#define  LP_WAKE_CTRL_5           0x02
#define  LP_WAKE_CTRL_10          0x03

#define MPU6050_REG_PWR_MGMT_2        (0x6C) // Power Management 2

void lowPowerAccel(uint8_t frequency) {
  uint8_t value;
  value = mpu.readRegister8(MPU6050_REG_PWR_MGMT_2);
  value &= 0b00111000;
  value |= (frequency << 6) | 0b111;
  mpu.writeRegister8(MPU6050_REG_PWR_MGMT_2, value);

  value = mpu.readRegister8(MPU6050_REG_PWR_MGMT_1);
  value &= 0b10010111;
  value |= 0b00111000;
  mpu.writeRegister8(MPU6050_REG_PWR_MGMT_1, value);
}


void init() {
  Serial.begin(115200);
  Serial.println("Initialize MPU6050");
  while(!mpu.begin(MPU6050_SCALE_2000DPS, MPU6050_RANGE_16G))
  {
    Serial.println("Could not find a valid MPU6050 sensor, check wiring!");
    delay(500);
  }

  mpu.setAccelPowerOnDelay(MPU6050_DELAY_3MS);

  mpu.setIntFreeFallEnabled(false);
  mpu.setIntZeroMotionEnabled(false);
  mpu.setIntMotionEnabled(true);

  mpu.setDHPFMode(MPU6050_DHPF_5HZ);

  mpu.setMotionDetectionThreshold(2);
  mpu.setMotionDetectionDuration(5);

  mpu.setZeroMotionDetectionThreshold(4);
  mpu.setZeroMotionDetectionDuration(2);

  checkSettings();

  #ifdef ENABLE_LEDS
    pinMode(PIN_LED1, OUTPUT);
    digitalWrite(PIN_LED1, LOW);

    pinMode(PIN_LED2, OUTPUT);
    digitalWrite(PIN_LED2, LOW);
  #endif

  #ifdef ENABLE_SOUND
    pinMode(PIN_SPEAKER, OUTPUT);
  #endif
  lowPowerAccel(LP_WAKE_CTRL_1_25);
}

#define GRAVITATIONAL_ACCELERATION 9.81
#define TIME_TO_GET_READY 10
#define ACC_STEP 10

void calibrate() {
  for (int i = 0; i < TIME_TO_GET_READY; i++) {
    delay(1000);
    Serial.println("Get ready to calibrate LEFT" + (TIME_TO_GET_READY - i));
  }
    
  for (int i = 32000; i > -32000; i = i - 10) {
    mpu.setAccelOffsetY(i);
    delay(1);
    if (mpu.readNormalizeAccel().YAxis < GRAVITATIONAL_ACCELERATION) break;
  }

  for (int i = 0; i < TIME_TO_GET_READY; i++) {
    delay(1000);
    Serial.println("Get ready to calibrate STAND" + (TIME_TO_GET_READY - i));
  }
    
  for (int i = 32000; i > -32000; i = i - 10) {
    mpu.setAccelOffsetX(i);
    delay(1);
    if (mpu.readNormalizeAccel().XAxis < GRAVITATIONAL_ACCELERATION) break;
  }

  for (int i = 0; i < TIME_TO_GET_READY; i++) {
    delay(1000);
    Serial.println("Get ready to calibrate SUPINE" + (TIME_TO_GET_READY - i));
  }
    
  for (int i = 32000; i > -32000; i = i - 10) {
    mpu.setAccelOffsetZ(i);
    delay(1);
    if (mpu.readNormalizeAccel().ZAxis < GRAVITATIONAL_ACCELERATION) break;
  }
}

void checkSettings()
{
  Serial.println();

  Serial.print(" * Sleep Mode:                ");
  Serial.println(mpu.getSleepEnabled() ? "Enabled" : "Disabled");

  Serial.print(" * Motion Interrupt:     ");
  Serial.println(mpu.getIntMotionEnabled() ? "Enabled" : "Disabled");

  Serial.print(" * Zero Motion Interrupt:     ");
  Serial.println(mpu.getIntZeroMotionEnabled() ? "Enabled" : "Disabled");

  Serial.print(" * Free Fall Interrupt:       ");
  Serial.println(mpu.getIntFreeFallEnabled() ? "Enabled" : "Disabled");

  Serial.print(" * Motion Threshold:          ");
  Serial.println(mpu.getMotionDetectionThreshold());

  Serial.print(" * Motion Duration:           ");
  Serial.println(mpu.getMotionDetectionDuration());

  Serial.print(" * Zero Motion Threshold:     ");
  Serial.println(mpu.getZeroMotionDetectionThreshold());

  Serial.print(" * Zero Motion Duration:      ");
  Serial.println(mpu.getZeroMotionDetectionDuration());

  Serial.print(" * Clock Source:              ");
  switch(mpu.getClockSource())
  {
    case MPU6050_CLOCK_KEEP_RESET:     Serial.println("Stops the clock and keeps the timing generator in reset"); break;
    case MPU6050_CLOCK_EXTERNAL_19MHZ: Serial.println("PLL with external 19.2MHz reference"); break;
    case MPU6050_CLOCK_EXTERNAL_32KHZ: Serial.println("PLL with external 32.768kHz reference"); break;
    case MPU6050_CLOCK_PLL_ZGYRO:      Serial.println("PLL with Z axis gyroscope reference"); break;
    case MPU6050_CLOCK_PLL_YGYRO:      Serial.println("PLL with Y axis gyroscope reference"); break;
    case MPU6050_CLOCK_PLL_XGYRO:      Serial.println("PLL with X axis gyroscope reference"); break;
    case MPU6050_CLOCK_INTERNAL_8MHZ:  Serial.println("Internal 8MHz oscillator"); break;
  }

  Serial.print(" * Accelerometer:             ");
  switch(mpu.getRange())
  {
    case MPU6050_RANGE_16G:            Serial.println("+/- 16 g"); break;
    case MPU6050_RANGE_8G:             Serial.println("+/- 8 g"); break;
    case MPU6050_RANGE_4G:             Serial.println("+/- 4 g"); break;
    case MPU6050_RANGE_2G:             Serial.println("+/- 2 g"); break;
  }

  Serial.print(" * Accelerometer offsets:     ");
  Serial.print(mpu.getAccelOffsetX());
  Serial.print(" / ");
  Serial.print(mpu.getAccelOffsetY());
  Serial.print(" / ");
  Serial.println(mpu.getAccelOffsetZ());

  Serial.print(" * Accelerometer power delay: ");
  switch(mpu.getAccelPowerOnDelay())
  {
    case MPU6050_DELAY_3MS:            Serial.println("3ms"); break;
    case MPU6050_DELAY_2MS:            Serial.println("2ms"); break;
    case MPU6050_DELAY_1MS:            Serial.println("1ms"); break;
    case MPU6050_NO_DELAY:             Serial.println("0ms"); break;
  }

  Serial.println();
}

Vector getAcc() {
  return lastrawAccel;
}

void loop()
{
  //checkSettings();
  lastrawAccel = mpu.readNormalizeAccel();

  /*
  Activites act = mpu.readActivites();

  if (act.isActivity)
  {
    #ifdef ENABLE_LEDS
      digitalWrite(PIN_LED1, HIGH);
      digitalWrite(PIN_LED2, HIGH);
    #endif

    #ifdef ENABLE_SOUND
      tone(PIN_SPEAKER, 1000, 100);
    #endif
  } else
  {
    #ifdef ENABLE_LEDS
      digitalWrite(PIN_LED1, LOW);
      digitalWrite(PIN_LED2, LOW);
    #endif
  }

  if (act.isInactivity)
  {
    //digitalWrite(6, HIGH);
  } else
  {
    //digitalWrite(6, LOW);
  }

  Serial.print(act.isActivity);
  Serial.print(act.isInactivity);

  Serial.print(" ");
  Serial.print(act.isPosActivityOnX);
  Serial.print(act.isNegActivityOnX);
  Serial.print(" ");

  Serial.print(act.isPosActivityOnY);
  Serial.print(act.isNegActivityOnY);
  Serial.print(" ");

  Serial.print(act.isPosActivityOnZ);
  Serial.print(act.isNegActivityOnZ);
  Serial.print("\n");
  */

  SERIAL_PRINT(" * Accelerometer offsets:     ");
  SERIAL_PRINT(lastrawAccel.XAxis);
  SERIAL_PRINT(" / ");
  SERIAL_PRINT(lastrawAccel.YAxis);
  SERIAL_PRINT(" / ");
  SERIAL_PRINTLN(lastrawAccel.ZAxis);

  SERIAL_PRINT(mpu.getAccelOffsetX());
  SERIAL_PRINT(" / ");
  SERIAL_PRINT(mpu.getAccelOffsetY());  
  SERIAL_PRINT(" / ");
  SERIAL_PRINTLN(mpu.getAccelOffsetZ());
  //delay(500);
} 
};
