/**
 * Copyright ISS2020
 * @Author: Calin Bira (calin.bira@spacescience.ro / calin.bira@upb.ro)
 */
 
#define LOW_POWER_MODE

#ifdef LOW_POWER_MODE
  #define SERIAL_PRINT(x)
  #define SERIAL_PRINTLN(x)
#else
  #define SERIAL_PRINT(x) Serial.print(x)
  #define SERIAL_PRINTLN(x) Serial.println(x)
#endif

#include <avr/sleep.h>
#include "iss_wx_button.hpp"
#include "iss_wx_led.hpp"
#include "iss_wx_display.hpp"
#include "iss_wx_ble.hpp"
#include "iss_wx_battery.hpp"
#include "iss_wx_accelerometer.hpp"
#include "iss_wx_bodyposition.hpp"
#include "iss_wx_watchdog.hpp"

issWxButton issbuttons;
issWxLed issleds;
issWxDisplay issdisplay;
issWxBLE issble;
issWxBattery issbattery;
issWxAccelerometer issacc;
issWxBodyposition issbodypos;

WatchdogAVR watchdog;

uint8_t oldBodyPosition;
uint8_t newBodyPosition;

void setup() {
  issbuttons.init();
  issleds.init();
  issdisplay.init();
  issble.init();
  issbattery.init(issble.getBLEhandle());
  issbattery.loop();
  issacc.init();
  //issacc.calibrate(); // do only once, and monitor output to make sure calibration was successful
  issbodypos.init();

  issleds.set(issWxLed::LEFT , issWxLed::OFF);
  issleds.set(issWxLed::RIGHT, issWxLed::OFF);

  issacc.loop();
  oldBodyPosition = issWxBodyposition::NON_DEFINED_POS;
  newBodyPosition = issbodypos.getPosition(issacc.getAcc());
}

enum {
  ROW_ISS_SAVE = 0,
  ROW_BODY_POSITION = 5,
  ROW_BATT_STATUS = 6,
  ROW_MINS_TO_END = 7,
}DisplayRows_t;

void loop() {
  issbattery.loop();
    
  static int loops = 0;
  static uint32_t lastms = millis();
  loops++;

  float soc = issbattery.getSOC();
  issdisplay.println("issSAVE 2020    " + String((int)soc) + "%", ROW_ISS_SAVE); 
  issdisplay.println(String((int)soc) + "% (" + issbattery.getMillivolts() + " mV)", ROW_BATT_STATUS);

  int16_t minutes = issbattery.getEnduranceMinutes();

  if (minutes < 0) 
    issdisplay.println(String(loops) + ":"+ String(-minutes) + " min to 100%", ROW_MINS_TO_END);  
  else if (minutes > 0)
    issdisplay.println(String(loops) + ":" + String(minutes) + " min to empty", ROW_MINS_TO_END);
  else 
    issdisplay.println(String(loops), ROW_MINS_TO_END);

    #ifdef LOW_POWER_MODE
      issdisplay.turn(true);
      if (issble.getBLEhandle()) {      
        issble.getBLEhandle()->print("AT+BLEUARTTX=");
        
        /*issble.getBLEhandle()->println(
            String(issbattery.getSOC())+"/"+
            String((int)issacc.getAcc().XAxis) + "/" + 
            String((int)issacc.getAcc().YAxis) + "/" + 
            String((int)issacc.getAcc().ZAxis));
        */
        
        String posOld = issbodypos.getPositionName(oldBodyPosition);
        String posNew = issbodypos.getPositionName(newBodyPosition);
        issble.getBLEhandle()->println(posOld + "=>" + posNew);
        issdisplay.println(posOld + " => " + posNew, ROW_BODY_POSITION);
        oldBodyPosition = newBodyPosition;
      }
      while ((millis() - lastms) < 10000)
      ; // wait x seconds in total before shutting screen down
      issdisplay.turn(false);
    
      while (true) {
        watchdog.sleep(200);
        issacc.loop();
        newBodyPosition = issbodypos.getPosition(issacc.getAcc());
        if (newBodyPosition != oldBodyPosition) break;
      }
    #else
      issacc.loop();
    #endif    
    lastms = millis();
}

// Define watchdog timer interrupt.
ISR(WDT_vect) {
  // Nothing needs to be done, however interrupt handler must be defined to prevent a reset.
}
