#pragma once

// Be careful to use a platform-specific conditional include to only make the
// code visible for the appropriate platform.  Arduino will try to compile and
// link all .cpp files regardless of platform.
#include <avr/interrupt.h>
#include <avr/power.h>
#include <avr/sleep.h>
#include <avr/wdt.h>

class WatchdogAVR {
public:
  WatchdogAVR() : _wdto(-1) {}

  // Enable the watchdog timer to reset the machine after a period of time
  // without any calls to reset().  The passed in period (in milliseconds) is
  // just a suggestion and a lower value might be picked if the hardware does
  // not support the exact desired value.
  
  
  // The actual period (in milliseconds) before a watchdog timer reset is returned.
  int enable(int maxPeriodMS = 0) {
    // Pick the closest appropriate watchdog timer value.
    int actualMS;
    _setPeriod(maxPeriodMS, _wdto, actualMS);
    // Enable the watchdog and return the actual countdown value.
    wdt_enable(_wdto);
    return actualMS;
  }

  // Reset or 'kick' the watchdog timer to prevent a reset of the device.
  void reset() {
    // Reset the watchdog.
    wdt_reset();
  }

  // Completely disable the watchdog timer.
  void disable() {
    // Disable the watchdog and clear any saved watchdog timer value.
    wdt_disable();
    _wdto = -1;
  }

  // Enter the lowest power sleep mode (using the watchdog timer) for the
  // desired period of time.  The passed in period (in milliseconds) is
  // just a suggestion and a lower value might be picked if the hardware does
  // not support the exact desired value
  //
  // The actual period (in milliseconds) that the hardware was asleep will be returned.
  int sleep(int maxPeriodMS = 0) {
    // Pick the closest appropriate watchdog timer value.
    int sleepWDTO, actualMS;
    _setPeriod(maxPeriodMS, sleepWDTO, actualMS);
  
    // Build watchdog prescaler register value before timing critical code.
    uint8_t wdps = ((sleepWDTO & 0x08 ? 1 : 0) << WDP3) |
                   ((sleepWDTO & 0x04 ? 1 : 0) << WDP2) |
                   ((sleepWDTO & 0x02 ? 1 : 0) << WDP1) |
                   ((sleepWDTO & 0x01 ? 1 : 0) << WDP0);
  
    // The next section is timing critical so interrupts are disabled.
    cli();
    // First clear any previous watchdog reset.
    MCUSR &= ~(1 << WDRF);
    // Now change the watchdog prescaler and interrupt enable bit so the
    // watchdog reset only triggers the interrupt (and wakes from deep sleep)
    // and not a full device reset.  This is a timing critical section of
    // code that must happen in 4 cycles.
    WDTCSR |= (1 << WDCE) | (1 << WDE); // Set WDCE and WDE to enable changes.
    WDTCSR = wdps;                      // Set the prescaler bit values.
    WDTCSR |= (1 << WDIE);              // Enable only watchdog interrupts.
    // Critical section finished, re-enable interrupts.
    sei();
  
    // Disable USB if it exists
  #ifdef USBCON
    USBCON |= _BV(FRZCLK); // freeze USB clock
    PLLCSR &= ~_BV(PLLE);  // turn off USB PLL
    USBCON &= ~_BV(USBE);  // disable USB
  #endif
  
    // Set full power-down sleep mode and go to sleep.
    set_sleep_mode(SLEEP_MODE_PWR_DOWN);
    sleep_mode();
  
    // Chip is now asleep!
  
    // Once awakened by the watchdog execution resumes here.
    // Start by disabling sleep.
    sleep_disable();
  
    // Check if user had the watchdog enabled before sleep and re-enable it.
    if (_wdto != -1)
      wdt_enable(_wdto);
  
    // Return how many actual milliseconds were spent sleeping.
    return actualMS;
  }

private:
  // Pick the closest (but not higher) watchdog timer value from the provided
  // maximum period.  Sets wdto to the chosen period value suitable for
  // passing to wdt_enable(), and actualMS to the chosen period value in
  // milliseconds.  A max value of 0 will pick the longest value possible.
  
  void _setPeriod(int maxMS, int &wdto, int &actualMS) {
    // Note the order of these if statements from highest to lowest  is
    // important so that control flow cascades down to the right value based
    // on its position in the range of discrete timeouts.
    if ((maxMS >= 8000) || (maxMS == 0)) {
      wdto = WDTO_8S;
      actualMS = 8000;
    } else if (maxMS >= 4000) {
      wdto = WDTO_4S;
      actualMS = 4000;
    } else if (maxMS >= 2000) {
      wdto = WDTO_2S;
      actualMS = 2000;
    } else if (maxMS >= 1000) {
      wdto = WDTO_1S;
      actualMS = 1000;
    } else if (maxMS >= 500) {
      wdto = WDTO_500MS;
      actualMS = 500;
    } else if (maxMS >= 250) {
      wdto = WDTO_250MS;
      actualMS = 250;
    } else if (maxMS >= 120) {
      wdto = WDTO_120MS;
      actualMS = 120;
    } else if (maxMS >= 60) {
      wdto = WDTO_60MS;
      actualMS = 60;
    } else if (maxMS >= 30) {
      wdto = WDTO_30MS;
      actualMS = 30;
    } else {
      wdto = WDTO_15MS;
      actualMS = 15;
    }
  }

  // Keep the last selected watchdog timer period so that the watchdog can be
  // re-enabled at that rate after sleep.  A value of -1 means no watchdog
  // timer was enabled.
  int _wdto;
};
