/**
 * ISS Copyright 2019-2020
 * @author Calin Bira (calin.bira@spacescience.ro / calin.bira@upb.ro)
 * @author Alexandru Mandu (alexandru.mandu@spacescience.ro)
 */

#pragma once
#include <Adafruit_GFX_AS.h>
#include <Adafruit_ILI9341_AS.h>

class issBodyScale : issModule {
    enum Gender {
      FEMALE_GENDER = 0,
      MALE_GENDER = 1
    };

    char* MAC_SCALE;

    //Scale User profile
    uint8_t user_gender = 1;   //1-Male    0-Female
    uint8_t user_height = 175; //in cm (0-255)
    uint8_t user_age = 29;     //(0-255)

    uint8_t available_scale = 0;
    uint8_t connected_scale = 0;
    uint8_t connection_handle_scale = 0;

    #define SCALE_HANDLE 83
    #define SCALE_PROFILE_HANDLE 85


    public:

        //!Struct to store data of the glucometer.
        struct scaleDataVector
        {
          uint16_t weight;
          uint16_t bodyfat;
          uint16_t musclemass;
          uint16_t water;
          uint16_t calories;
          uint8_t visceralfat;
          uint8_t bonemass;
        };

  private:
      //!Vector to store the glucometer measures and dates.
      scaleDataVector mBodyScaleResults;
      issBLE *pBLE;

  public:
        issBodyScale(char *mac) {
          MAC_SCALE = mac;
        }

        char* getMac() {
          return MAC_SCALE;
        }

        void init(issDisplay *_pDisplay, issBLE *_pBLE)
        {
            pBLE = _pBLE;
            pDisplay = _pDisplay;

            if (! pBLE->isInitialized())
              pBLE->init(_pDisplay);
            pBLE->pushMAC(MAC_SCALE);

            mInitialized = true;

            mBodyScaleResults.weight = 0;
            mBodyScaleResults.bodyfat = 0;
            mBodyScaleResults.musclemass = 0;
            mBodyScaleResults.water = 0;
            mBodyScaleResults.calories = 0;
            mBodyScaleResults.visceralfat = 0;
            mBodyScaleResults.bonemass = 0;
        }

        void setParams(uint8_t gender_01, uint8_t height_cm, uint8_t age_year){
              user_gender = gender_01;   //1-Male    0-Female
              user_height = height_cm; //in cm (0-255)
              user_age = age_year;     //(0-255)
        }

        void loop()
        {
          #ifndef USE_DUMMY_DATA
          if (pBLE) pBLE->reset();
          available_scale = MySignals_BLE.scanDevice(MAC_SCALE, 1000, TX_POWER_MAX);

          uint16_t y = pDisplay->getNextY();
          //if (pDisplay != NULL) pDisplay->drawString("Scale available:", 0, y, pDisplay->getFontSize());
          //if (pDisplay != NULL) pDisplay->drawNumber(available_scale, 110, y, pDisplay->getFontSize());

          if (available_scale == 1)
          {
            if (MySignals_BLE.connectDirect(MAC_SCALE) == 1)
            {
              //if (pDisplay != NULL) pDisplay->drawString("Connected    ", 0, pDisplay->getNextY(), pDisplay->getFontSize());

              connected_scale = 1;
              delay(500);

              uint8_t attributeData[2] = { 0x01 , 0x00 };
              if (MySignals_BLE.attributeWrite(MySignals_BLE.connection_handle, SCALE_HANDLE, attributeData, 2) == 0)
              {
                //if (pDisplay != NULL) pDisplay->drawString("Subscribed", 0, pDisplay->getNextY(), pDisplay->getFontSize());
                delay(4000);

                uint8_t scale_user_profile[8] =
                {
                  0xfe,  // 0  BYTE 1
                  0x00,  // 1  BYTE 2 -> user group
                  0x00,  // 2  BYTE 3 -> gender
                  0x00,  // 3  BYTE 4 -> level
                  0x00,  // 4  BYTE 5 -> height
                  0x00,  // 5  BYTE 6 -> age
                  0x00,  // 6  BYTE 7 -> unit
                  0x00   // 7  BYTE 8 -> xor
                };

                scale_user_profile[1] = 1;              // User group
                scale_user_profile[2] = user_gender;    // gender: 1=male, 0=female
                scale_user_profile[3] = 0;              // level 0=normal
                scale_user_profile[4] = user_height;    // height
                scale_user_profile[5] = user_age;       // age
                scale_user_profile[6] = 1;              // unit KG

                uint8_t xor_result = scale_user_profile[1] xor scale_user_profile[2];
                xor_result = xor_result xor scale_user_profile[3];
                xor_result = xor_result xor scale_user_profile[4];
                xor_result = xor_result xor scale_user_profile[5];
                scale_user_profile[7] = xor_result xor scale_user_profile[6];

                if (MySignals_BLE.attributeWrite(MySignals_BLE.connection_handle, SCALE_PROFILE_HANDLE, scale_user_profile, 8) == 0)
                {
                  delay(500);
                  unsigned long previous = millis();
                  do
                  {
                    if (MySignals_BLE.waitEvent(1000) == BLE_EVENT_ATTCLIENT_ATTRIBUTE_VALUE)
                    {
                      uint8_t scale_weight_high = MySignals_BLE.event[13];
                      uint8_t scale_weight_low = MySignals_BLE.event[14];

                      uint8_t scale_bodyfat_high = MySignals_BLE.event[15];
                      uint8_t scale_bodyfat_low = MySignals_BLE.event[16];

                      uint8_t scale_musclemass_high = MySignals_BLE.event[18];
                      uint8_t scale_musclemass_low = MySignals_BLE.event[19];

                      mBodyScaleResults.visceralfat = MySignals_BLE.event[20];

                      uint8_t scale_water_high = MySignals_BLE.event[21];
                      uint8_t scale_water_low = MySignals_BLE.event[22];

                      uint8_t scale_calories_high = MySignals_BLE.event[23];
                      uint8_t scale_calories_low = MySignals_BLE.event[24];

                      mBodyScaleResults.weight = (scale_weight_high * 256) + scale_weight_low;
                      mBodyScaleResults.bodyfat = (scale_bodyfat_high * 256) + scale_bodyfat_low;
                      mBodyScaleResults.musclemass = (scale_musclemass_high * 256) + scale_musclemass_low;
                      mBodyScaleResults.water = (scale_water_high * 256) + scale_water_low;
                      mBodyScaleResults.calories = (scale_calories_high * 256) + scale_calories_low;

                      mBodyScaleResults.bonemass = MySignals_BLE.event[17] * 1000 / mBodyScaleResults.weight;
                      produce();
                      produce();
                      //show();

                      delay(1000);

                      uint8_t shutdown_command[8] = { 0xfd, 0x35, 0x00, 0x00, 0x00, 0x00, 0x00, 0x35 };
                      MySignals_BLE.attributeWrite(MySignals_BLE.connection_handle, SCALE_PROFILE_HANDLE, shutdown_command, 8);
                      MySignals_BLE.disconnect(MySignals_BLE.connection_handle);
                      delay(200);
                      connected_scale = 0;
                    }
                  }
                  while ((connected_scale == 1) && ((millis() - previous) < 20000));
                  connected_scale = 0;
                }
                else
                {
                    //if (pDisplay != NULL) pDisplay->drawString("Error subscribing", 0, pDisplay->getNextY(), pDisplay->getFontSize());
                }
              }
              else
              {
                  //if (pDisplay != NULL) pDisplay->drawString("Not Connected", 0, pDisplay->getNextY(), pDisplay->getFontSize());
              }
            }
            else
            {
              connected_scale = 0;
              //if (pDisplay != NULL) pDisplay->drawString("Not Connected", 0, pDisplay->getNextY(), pDisplay->getFontSize());
            }
          }
          else if (available_scale == 0)
          {
            //Do nothing
          }
          else
          {
            MySignals_BLE.hardwareReset();
            MySignals_BLE.initialize_BLE_values();
            delay(100);
          }
          //delay(1000);
         #else
         //  "SCALE": [ SUTE_DE_GRAME,  BODY_FAT, MUSCLE_MASS, WATER, CALORIES, BONE_MASS, VISCERAL_FAT ],
         // SUTE_DE_GRAME un numar int 0-9999 cu sensul de 0-999 kg, BODY_FAT, MUSCLE_MASS, WATER, CALORIES, BONE_MASS, VISCERAL_FAT sunt in procente

            mBodyScaleResults.weight = (mBodyScaleResults.weight+1) % 1000;
            mBodyScaleResults.bodyfat = (mBodyScaleResults.bodyfat + 1) % 101;
            mBodyScaleResults.musclemass = (mBodyScaleResults.musclemass + 1) % 101;
            mBodyScaleResults.water = (mBodyScaleResults.water + 1) % 101;
            mBodyScaleResults.calories = (mBodyScaleResults.calories + 1) % 101;
            mBodyScaleResults.visceralfat = (mBodyScaleResults.visceralfat + 1) % 101;
            mBodyScaleResults.bonemass = (mBodyScaleResults.bodyfat + 1) % 101;
            produce();
            produce();
         #endif
        }

    String consumeToJSONstring() {
      //"SCALE": [ SUTE_DE_GRAME,  BODY_FAT, MUSCLE_MASS, WATER, CALORIES, BONE_MASS, VISCERAL_FAT ]
      if (isAvailable()) {consume(); return ",\"SCALE\":[" +
                            String(mBodyScaleResults.weight*10) +  "," +
                            String(mBodyScaleResults.bodyfat) +    "," +
                            String(mBodyScaleResults.musclemass) + "," +
                            String(mBodyScaleResults.water) +      "," +
                            String(mBodyScaleResults.calories) +   "," +
                            String(mBodyScaleResults.bonemass) +   "," +
                            String(mBodyScaleResults.visceralfat)+ "]";}
      else return String("");
    }

    String consumeToDisplayedString() {
      if (isAvailable()) {consume(); return " SCALE " + String(mBodyScaleResults.weight / 10) + "," + String(mBodyScaleResults.weight % 10) + " kg" + getEmptyString();}
      else return " SCALE" + getEmptyString();
    }

    void show() {
      if (pDisplay != NULL) {
          if (!isAvailable()) {pDisplay->drawString("Bodyscale: n/a"); return;}

          int y = pDisplay->getNextY();

          pDisplay->drawString("Weight:", 0, y, pDisplay->getFontSize());
          pDisplay->drawFloat((mBodyScaleResults.weight / 10.0), 1, 100, y, pDisplay->getFontSize());

          /*
          y = pDisplay->getNextY();
          pDisplay->drawString("Body fat:", 0, y, pDisplay->getFontSize());
          pDisplay->drawFloat((mBodyScaleResults.bodyfat / 10.0), 1, 100, y, pDisplay->getFontSize());

          y = pDisplay->getNextY();
          pDisplay->drawString("Muscle mass:", 0, y, pDisplay->getFontSize());
          pDisplay->drawFloat((mBodyScaleResults.musclemass / 10.0), 1, 100, y, pDisplay->getFontSize());

          y = pDisplay->getNextY();
          pDisplay->drawString("Visceral fat:", 0, y, pDisplay->getFontSize());
          pDisplay->drawNumber(mBodyScaleResults.visceralfat, 100, y, pDisplay->getFontSize());

          y = pDisplay->getNextY();
          pDisplay->drawString("Water:", 0, y, pDisplay->getFontSize());
          pDisplay->drawFloat((mBodyScaleResults.water / 10.0), 1, 100, y, pDisplay->getFontSize());

          y = pDisplay->getNextY();
          pDisplay->drawString("Calories:", 0, y, pDisplay->getFontSize());
          pDisplay->drawNumber(mBodyScaleResults.calories, 100, y, pDisplay->getFontSize());

          y = pDisplay->getNextY();
          pDisplay->drawString("Bone mass:", 0, y, pDisplay->getFontSize());
          pDisplay->drawFloat((mBodyScaleResults.bonemass / 10.0), 1, 100, y, pDisplay->getFontSize());
          */
        }
    }
};
