/**
 * ISS Copyright 2020
 * @author Calin Bira (calin.bira@spacescience.ro / calin.bira@upb.ro)
 */

#pragma once

#include "issModule.hpp"

class issTemperature : public issModule {
  uint16_t lastValidValue = 0;

  public:
    void show() {
      if (pDisplay != NULL) {
        if (isAvailable())
          pDisplay->drawString((String("Temperature (deg. C) : ") + String(lastValidValue/10)).c_str());
        else
          pDisplay->drawString((String("Temperature (deg. C) : n/a")).c_str());
      }
  }

  float getTempC() {
    return getCalibratedTemperature(100,10,0.2,5);
  }

  void loop() {
    // "TEMP": [T_SUTIMI_DE_GRADE_CELSIUS], // int 0-999 ( echivalent 0 - 99.9 grade Celsius)

    #ifndef USE_DUMMY_DATA
      float lastValue = getCalibratedTemperature(100,10,0.2,5);
      if (lastValue == 0) return;
      //else:
      lastValidValue = lastValue*10;
    #else
      lastValidValue = ((lastValidValue + 1) % 100) + 300;
    #endif
    produce();
    produce();
  }

  String consumeToDisplayedString() {
    if (isAvailable()) {consume(); return " TEMPERATURE " + String(lastValidValue / 10) + "," + String(lastValidValue % 10) + " deg. C" + getEmptyString();}
    else return " TEMPERATURE" + getEmptyString();
  }

  String consumeToJSONstring() {
    if (isAvailable()) {consume(); return ",\"TEMP\":[" + String(lastValidValue) + "]";}
    else return "";
  }

 private:
  float getCalibratedTemperature(uint8_t samples, uint16_t delaySample, float offset , uint8_t format)
  {
    //Local variables
    float temperature; //Corporal Temperature
    float resistance;  //Resistance of MySignals.
    float ganancia = 5.0;
    float Vcc = 5.0;
    float refTension = 3.0; // Voltage Reference of Wheatstone bridge.
    float Ra = 4700.0; //Wheatstone bridge resistance.
    float Rc = 4700.0; //Wheatstone bridge resistance.
    float Rb = 820.0; //Wheatstone bridge resistance.
    float samplesArray[samples];
    float voltage;

    for (int i = 0; i < samples; i++)
    {
      // disable global interrupts
      //cli();
      //int sensorValue = MySignals.readADC(ADC_TEMPERATURE); //don't use, because the ADC located on MySignals board (MCP3208) does not work properly
        int sensorValue = analogRead(A6); // use, because the ADC located on Arduino MEGA2560 board (via A6 pin) work properly
      // enable global interrupts
      //sei();

      // binary to voltage conversion
      voltage = ((float)sensorValue * Vcc) / 1023;
      // Wheatstone bridge output voltage.
      voltage = voltage / ganancia;
      // Resistance sensor calculate
      float aux = (voltage / refTension) + Rb / (Rb + Ra);
      resistance = Rc * aux / (1 - aux);

      delay(delaySample);

      //DEBUG_SERIAL.println(resistance);
      if (resistance >= 1822.8) // under 30 degrees
      {
        // if temperature between 25ºC and 29.9ºC. R(tª)=6638.20457*(0.95768)^t
        //samplesArray[i] = log(resistance / 6638.20457) / log(0.95768);
        samplesArray[i] = offset;
      }
      else if (resistance >= 1477.1)
      {
        // if temperature between 30ºC and 34.9ºC. R(tª)=6403.49306*(0.95883)^t
        samplesArray[i] = log(resistance / 6403.49306) / log(0.95883);
      }
      else if (resistance >= 1204.8)
      {
        // if temperature between 35ºC and 39.9ºC. R(tª)=6118.01620*(0.96008)^t
        samplesArray[i] = log(resistance / 6118.01620) / log(0.96008);
      }
      else if (resistance >= 988.1)
      {
        // if temperature between 40ºC and 44.9ºC. R(tª)=5859.06368*(0.96112)^t
        samplesArray[i] = log(resistance / 5859.06368) / log(0.96112);
      }
      else if (resistance >= 811.7)
      {
        // if temperature between 45ºC and 50ºC. R(tª)=5575.94572*(0.96218)^t
        samplesArray[i] = log(resistance / 5575.94572) / log(0.96218);
      }
    }

    temperature = median(samplesArray, samples);
    temperature = temperature - offset;

  #if MYHOSPITAL_DEBUG > 0
    Serial.println();
    Serial.println("-----------------> TEMPERATURE SENSOR");
    Serial.print("Calibrated temperature = ");
    Serial.println(temperature);
    Serial.print("Resistance = ");
    Serial.println(resistance);
    Serial.print("Voltage = ");
    Serial.println(voltage);
  #endif

    if (format == 5) return temperature;
    if (format == 3) return voltage;
    else return resistance;
  }

  float median(float* values, uint8_t length) {
    float man;
    bool is_ordered;
    do {
          is_ordered = true;
          for (uint8_t i = 0; i < length-1; i++)
            if (values[i] > values[i+1]) {
              man = values[i];
              values[i] = values[i+1];
              values[i+1] = man;
              is_ordered = false;
            }
    } while (false == is_ordered);
    return values[length / 2];
  }
};
