/**
 * This file is part of Agni Testing Environment.
 * <p>
 * Agni Testing Environment is free software: you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * <p>
 * Agni Testing Environment is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 * <p>
 * You should have received a copy of the GNU General Public License along with Agni Testing Environment. If not, see
 * <http://www.gnu.org/licenses/>.
 * <p>
 * Copyright 2014 Radu Hobincu
 */
package utils;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.logging.Level;

/**
 * System logger.
 *
 * @author Radu Hobincu &lt;radu.hobincu@upb.ro&gt;
 */
public class Logger {

    /**
     * Marker for blue console output text.
     */
    private static final String ANSI_GREEN = "\u001B[32m";

    /**
     * Marker for blue console output text.
     */
    private static final String ANSI_BLUE = "\u001B[34m";

    /**
     * Marker for red console output text.
     */
    private static final String ANSI_RED = "\u001B[31m";

    /**
     * Marker for resetting the console output text color.
     */
    private static final String ANSI_RESET = "\u001B[0m";

    /**
     * Marker for yellow console output text.
     */
    private static final String ANSI_YELLOW = "\u001B[33m";

    /**
     * The current log level. Can be changed by editing the server.properties file.
     */
    private static Level sCurrentLevel = Level.ALL;

    /**
     * The output stream where the logging should be directed to.
     */
    private static PrintStream sPrintStream = System.out;

    /**
     * Object used to format the date and time.
     */
    private static final SimpleDateFormat FORMAT = new SimpleDateFormat(
            //"HH:mm:ss.SSS dd/MM/yy");
            "HH:mm:ss");

    /**
     * Current message index, it is incremented with every log line.
     */
    private static long sCurrentMessage = 0;

    /**
     * Sets the current log level to the specified level.
     *
     * @param level the level to set the logger to.
     */
    public synchronized static void setLogLevel(Level level) {
        sCurrentLevel = level;
        //log(level, "Log level set to: " + level);
    }

    /**
     * Returns the current log level to the specified level.
     *
     * @return the level the logger is set to.
     */
    public static Level getLogLevel() {
        return sCurrentLevel;
    }

    /**
     * Sets the name of the file in which the logger should output the messages.
     *
     * @param file The name of the file in which the logger should output the messages.
     */
    public synchronized static void setLogFilename(File file) {
        log(Level.INFO, "Redirected logging to: " + file);
        if (file != null && !file.isDirectory()) {
            try {
                sPrintStream = new PrintStream(file);
                log(Level.INFO, "Logging successfully redirected to " + file);
            } catch (FileNotFoundException ex) {
                Logger.log(Level.SEVERE, "", ex);
            }
        } else {
            sPrintStream = System.out;
            if (file != null) {
                log(Level.SEVERE, "Logging redirection failed to ", file);
            }
        }
    }

    /**
     * Method used to log a message and an exception stack trace.
     *
     * @param level     The log level of this message.
     * @param throwable The exception to log.
     * @param message   The message to log.
     * @param args      String arguments.
     *
     */
    public static void log(Level level, Throwable throwable, String message, Object... args) {
        if (level.intValue() < sCurrentLevel.intValue()) {
            return;
        }

        StackTraceElement element = Thread.currentThread().getStackTrace()[2];

        StringBuilder builder = new StringBuilder();
        if (level == Level.OFF) {
            builder.append(ANSI_GREEN);
        } else if (level == Level.SEVERE) {
            builder.append(ANSI_RED);
        } else if (level == Level.WARNING) {
            builder.append(ANSI_YELLOW);
        } else if (level != Level.INFO) {
            builder.append(ANSI_BLUE);
        }

        builder.append(buildMessage(element, level, String.format(message, args)));
        synchronized (Logger.class) {
            sPrintStream.println(builder.toString());
            if (sPrintStream != System.out) {
                System.out.println(builder.toString());
            }
            if (throwable != null) {
                throwable.printStackTrace(sPrintStream);
                if (sPrintStream != System.out) {
                    throwable.printStackTrace(System.out);
                }
            }
            sPrintStream.print(ANSI_RESET);
            if (sPrintStream != System.out) {
                System.out.print(ANSI_RESET);
            }
        }
    }

    /**
     * Method used to log a message with arguments.
     *
     * @param level   The log level of this message.
     * @param message The message to log.
     * @param args    The arguments for the message.
     */
    public static void log(Level level, String message, Object... args) {
        if (level.intValue() < sCurrentLevel.intValue()) {
            return;
        }

        StackTraceElement element = Thread.currentThread().getStackTrace()[2];

        StringBuilder builder = new StringBuilder();
        if (level == Level.SEVERE) {
            builder.append(ANSI_RED);
        } else if (level == Level.WARNING) {
            builder.append(ANSI_YELLOW);
        } else if (level != Level.INFO) {
            builder.append(ANSI_BLUE);
        }

        String str = buildMessage(element, level, String.format(message, args));
        builder.append(str);

        synchronized (Logger.class) {
            sPrintStream.println(builder.toString());
            sPrintStream.print(ANSI_RESET);

            if (sPrintStream != System.out) {
                System.out.println(builder.toString());
            }
            if (sPrintStream != System.out) {
                System.out.print(ANSI_RESET);
            }
        }
    }

    /**
     * Method used to format the message according to the log format.
     *
     * @param stackTraceElement The stack trace element for the calling method. Used to extract the class, method and
     *                          line where the log method was called.
     * @param level             The level of this log message.
     * @param message           The message to log.
     *
     * @return The constructed log message line.
     */
    private static String buildMessage(StackTraceElement stackTraceElement,
            Level level, String message) {
        StringBuilder builder = new StringBuilder();
        builder.append('[');
        builder.append(FORMAT.format(new Date(System.currentTimeMillis())));
        builder.append("] [");
        builder.append(String.format("%5d][", ++sCurrentMessage));
        if (level == Level.SEVERE) {
            builder.append("C][");
        } else if (level == Level.WARNING) {
            builder.append("W][");
        } else if (level == Level.INFO) {
            builder.append("I][");
        } else {
            builder.append("T][");
        }
        builder.append(Thread.currentThread().getName());
        builder.append("] [");
        builder.append(stackTraceElement.getClassName());
        builder.append('.');
        builder.append(stackTraceElement.getMethodName());
        builder.append(':');
        builder.append(stackTraceElement.getLineNumber());
        builder.append("]: ");
        builder.append(message);
        return builder.toString();
    }
}
