/**

Copyright 2020
Author: Calin Bira (calin.bira@upb.ro)

*/
#include <iostream>
#include <cstring>
#include <cstdint>
#include <fstream>

#define ErrorLine printf
#define WarningLine printf

#define PILRC_BYTE uint8_t
#define PILRC_USHORT uint16_t
#define PILRC_ULONG uint32_t

typedef struct BITMAPFILEHEADER
{
    PILRC_USHORT bfType;
    PILRC_USHORT bfSize1;
    PILRC_USHORT bfSize2;
    PILRC_USHORT bfReserved1;
    PILRC_USHORT bfReserved2;
    PILRC_USHORT bfOffBits1;
    PILRC_USHORT bfOffBits2;
} BITMAPFILEHEADER;

typedef struct BITMAPINFOHEADER
{
    PILRC_ULONG biSize;
    PILRC_ULONG biWidth;
    PILRC_ULONG biHeight;
    PILRC_USHORT biPlanes;
    PILRC_USHORT biBitCount;
    PILRC_ULONG biCompression;
    PILRC_ULONG biSizeImage;
    PILRC_ULONG biXPelsPerMeter;
    PILRC_ULONG biYPelsPerMeter;
    PILRC_ULONG biClrUsed;
    PILRC_ULONG biClrImportant;
} BITMAPINFOHEADER;

typedef struct RGBQUAD
{
    PILRC_BYTE rgbBlue;
    PILRC_BYTE rgbGreen;
    PILRC_BYTE rgbRed;
    PILRC_BYTE rgbReserved;
} RGBQUAD;

typedef struct BITMAPINFO
{
    BITMAPINFOHEADER bmiHeader;
    RGBQUAD bmiColors[1]; /* size is indefinite */
} BITMAPINFO;

/**
 *
 * @param fileName   the source file name
 */
void BitmapToCpp(const char *fileName)
{
    uint8_t *pBMPData;
    const char *pchExt;
    FILE *pFile;
    long size;

    // determine the size of the resource to load
    pFile = fopen(fileName, "rb");
    if (pFile == NULL) {
        ErrorLine("Could not find Resource %s.", fileName);
        return;
    }

    fseek(pFile, 0, SEEK_END);
    size = ftell(pFile);
    rewind(pFile);

    // alocate memory and load the resource in
    pBMPData = (uint8_t*)malloc(size);
    if (pBMPData == NULL) {
        ErrorLine("Resource %s too big to fit in memory", fileName);
        return;
    }
    fread(pBMPData, 1, size, pFile);
    fclose(pFile);

    BITMAPFILEHEADER *pbmf = (BITMAPFILEHEADER *)pBMPData;
    if (pbmf->bfType != 'M'*256 + 'B') {
        ErrorLine("Unknown bitmap type %hu", pbmf->bfType);
    }
    else {
        do {
                BITMAPINFO *pbmi = (BITMAPINFO *) (pBMPData + sizeof(BITMAPFILEHEADER));
                if (pbmi->bmiHeader.biBitCount != 8) {
                    ErrorLine("Not supporting bitmap type other than 8-bit color (with 4 unique colors)");
                    break;
                }

                RGBQUAD *rgbquad = (RGBQUAD *) (pBMPData + sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER));
                for (int i = 4; i < 256; i++)
                    if (((rgbquad + i)->rgbRed != 0) ||
                        ((rgbquad + i)->rgbGreen != 0) ||
                        ((rgbquad + i)->rgbBlue != 0)) {
                        ErrorLine("Not supporting bitmap type: other than 8-bit (with ONLY 4 unique colors)");
                        break;
                    }

                // 'logoiss8b', 320x240px
                //const uint16_t myBitmap [76800UL] PROGMEM = {
                std::ofstream outputhppfile;
                outputhppfile.open("array.hpp");
                outputhppfile << "// array of file: " << fileName << std::endl;
                outputhppfile << "// resolution of : " << pbmi->bmiHeader.biWidth << " x " << pbmi->bmiHeader.biHeight
                              << std::endl;
                outputhppfile << "// palette in BGR format" << std::endl << std::endl;

                outputhppfile << "uint16_t BitmapWidth = " << pbmi->bmiHeader.biWidth << ";" << std::endl;
                outputhppfile << "uint8_t BitmapHeight = " << pbmi->bmiHeader.biHeight << ";" << std::endl;

                outputhppfile << std::endl;
                outputhppfile << "const uint8_t BitmapPalette[256*3] PROGMEM = { " << std::endl;

                for (int i = 0; i < 256; i++) {
                    outputhppfile << (int) ((rgbquad + i)->rgbBlue) << "," <<
                                  (int) ((rgbquad + i)->rgbGreen) << "," <<
                                  (int) ((rgbquad + i)->rgbRed) << ",";

                    if ((i & 0x3F) == 0x3F) outputhppfile << std::endl;
                }

                outputhppfile << "};" << std::endl << std::endl;
                uint8_t *pixels = pBMPData + sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + 1024;

                outputhppfile << "const uint8_t BitmapPixels[" <<
                              pbmi->bmiHeader.biWidth * pbmi->bmiHeader.biHeight / 4 << "] PROGMEM = { " << std::endl;
                for (int y = 0; y < pbmi->bmiHeader.biHeight; y++) {
                    for (int x = 0; x < pbmi->bmiHeader.biWidth; x += 4) {
                        int value1 = (int) *(pixels + (pbmi->bmiHeader.biHeight - y - 1) * pbmi->bmiHeader.biWidth + x);
                        int value2 = (int) *(pixels + (pbmi->bmiHeader.biHeight - y - 1) * pbmi->bmiHeader.biWidth + x +
                                             1);
                        int value3 = (int) *(pixels + (pbmi->bmiHeader.biHeight - y - 1) * pbmi->bmiHeader.biWidth + x +
                                             2);
                        int value4 = (int) *(pixels + (pbmi->bmiHeader.biHeight - y - 1) * pbmi->bmiHeader.biWidth + x +
                                             3);

                        outputhppfile << (value1 << 6) + (value2 << 4) + (value3 << 2) + value4 << ",";
                    }
                    outputhppfile << std::endl;
                }
                outputhppfile << "};" << std::endl;
                outputhppfile.close();
                std::cout << "File converted successfully to array.hpp" << std::endl;
            } while(0);
        }

    // clean up
    free(pBMPData);
}

int main(int argc, char* argv[]) {
    std::string filename;
    if (argc == 1) {
        std::cout << "Please enter name of file to convert to 8-bit "
                     "color C++ array" << std::endl;
        std::cin >> filename;
    } else
        filename = std::string(argv[1]);

    BitmapToCpp(filename.c_str());
    //BmpImg logo;
    //logo.read(filename);
    //std::cout << "Image resolution is: " << logo.get_width() << " x " << logo.get_height() << std::endl;
    //std::cout << "Image bpp is " << logo.get_bpp() << std::endl;

    return 0;
}
