#include <iostream>
#include <filesystem>
#include <fstream>
#include <map>
#include "Exception.hpp"

using namespace std;

typedef struct  {
    string Name;
    int usStart;
    int usDuration;
    double IccAvg;
    double IccMin;
    double IccMax;
}EnergyElement_t;


typedef struct  {
    double Timestamp;
    double icc;
}IccElement_t;

class EnergyExtractor {

    double usAcquisitionPeriod = 0;
    double usEnergyTriggerTimestamp = 0;
    double CurrentTriggerTimestamp = 0;
    std::vector<EnergyElement_t> mEElements;
    std::vector<IccElement_t> mIElements;
public:

    bool doesFileExist(string name) {
        return filesystem::exists(name);
    }

    bool endsWith(string s, char end) {
        if (s.size() > 1)
            return s[s.size()-1] != end;
        return false;
    }

    bool pushTimings(string basename) {
        const int minLineSize = 8;
        // one log file, push name and time
        ifstream iflog(basename + ".log");
        if (!iflog.is_open()) throw Exception(".log file cannot be open");
        else
        {
            string line;
            int microsecondsStart;
            int microsecondsStop;

            while (getline(iflog, line))
            {
                stringstream ss(line);
                if (line.size() < minLineSize) break;

                string localTimestamp; ss >> localTimestamp;
                string marker; ss >> marker;

                string functionName;
                ss >> functionName;
                while (endsWith(functionName, ':')) {
                    string man;
                    ss >> man;
                    functionName += man;
                }

                ss >> microsecondsStart;
                string unit; ss >> unit;
                string to; ss >> to;

                ss >> microsecondsStop;
                cout << "(F)" << functionName << " " << microsecondsStop - microsecondsStart << endl;

                EnergyElement_t element;
                element.Name = functionName;
                element.usStart = microsecondsStart;
                element.usDuration = microsecondsStop - microsecondsStart;
                element.IccMin = 0;
                element.IccAvg = 0;
                element.IccMax = 0;
                mEElements.push_back(element);
            }
            iflog.close();

            // last event was the trigger
            usEnergyTriggerTimestamp = microsecondsStop;
        }

        return true;
    }

    bool pushCurrents(string basename) {
        const int minLineSize = 8;
        // one log file, push name and time
        ifstream iflog(basename + ".DAT");
        if (!iflog.is_open()) throw Exception(".DAT file cannot be open");
        else
        {
            string line;
            getline(iflog, line);
            stringstream ss(line);

            char comma;
            double oldlocalTimestamp; ss >> oldlocalTimestamp;
            ss >> comma;
            double oldtriggerData; ss >> oldtriggerData;
            ss >> comma;
            double oldcurrentData; ss >> oldcurrentData;

            usAcquisitionPeriod = 0.001; // 1GSa/s

            while (getline(iflog, line))
            {
                stringstream ss(line);
                if (line.size() < minLineSize) break;
                IccElement_t ie;

                double localTimestamp; ss >> localTimestamp;
                ie.Timestamp = localTimestamp;
                ss >> comma;

                double triggerData; ss >> triggerData;
                ss >> comma;

                double currentData; ss >> currentData;

                if ((oldtriggerData > 2.5) && (triggerData < 2.5)) // falling edge
                    CurrentTriggerTimestamp = localTimestamp;

                oldtriggerData = triggerData;
                ie.icc = currentData;
                mIElements.push_back(ie);
                //cout << "(mIElements)" << localTimestamp << " " << currentData << endl;
            }
            iflog.close();
            return true;
        }
    }

    void alignEnergyData() {
        const double SHUNT_RESISTOR = 0.925;
        cout << usAcquisitionPeriod << " us for period" << endl;
        cout << usEnergyTriggerTimestamp << " us for trigger-E" << endl;
        cout << CurrentTriggerTimestamp << " for trigger-I" << endl;

        int TriggerIndex;
        for (TriggerIndex = mIElements.size()-1; TriggerIndex >= 0; TriggerIndex--) {
            if (mIElements[TriggerIndex].Timestamp == CurrentTriggerTimestamp) {
                cout << mIElements[TriggerIndex].Timestamp << " " << CurrentTriggerTimestamp << endl;
                break;
            }
        }

        double Index = TriggerIndex;
        for (int i = mEElements.size() - 1; i >= 0; i--) {
            int samples = mEElements[i].usDuration * 1000; // 1000 samples per microsecond

            mEElements[i].IccMin = mIElements[Index].icc;
            mEElements[i].IccMax = mIElements[Index].icc;
            mEElements[i].IccAvg = 0;

            if (samples > Index) samples = Index;
            for (int k = Index - samples; k < Index; k++) {
                if (mEElements[i].IccMin > mIElements[k].icc)
                    mEElements[i].IccMin = mIElements[k].icc;

                if (mEElements[i].IccMax < mIElements[k].icc)
                    mEElements[i].IccMax = mIElements[k].icc;

                mEElements[i].IccAvg += mIElements[k].icc;
            }
            mEElements[i].IccAvg /= samples;
            mEElements[i].IccAvg /= SHUNT_RESISTOR;
            mEElements[i].IccMax /= SHUNT_RESISTOR;
            mEElements[i].IccMin /= SHUNT_RESISTOR;

            Index -= samples;
            cout << mEElements[i].Name << " " << mEElements[i].IccAvg*1000 << " mA ( "
            << mEElements[i].IccMin*1000 << " , " << mEElements[i].IccMax*1000 << " )"
                << Index << " samples:" << samples << "Energy "
                << mEElements[i].IccAvg * 5 * samples * 3.6 << endl;


            if (Index == 0) break;
        }
    }

    bool pushEnergy(string basename) {

        if (!doesFileExist(basename + ".log")) throw Exception(".log file does not exist");
        if (!doesFileExist(basename + ".DAT")) throw Exception(".DAT file does not exist");
        //if (doesFileExist(basename + ".png")) return false;

        bool ret = pushTimings(basename);
        ret &= pushCurrents(basename);
        alignEnergyData();

        return ret;
    }
};

int main() {
    EnergyExtractor ee;
    ee.pushEnergy("../logs/SDS00012");
    //ee.pushEnergy("../logs/SDS00015");
    //ee.pushEnergy("../logs/SDS00015");
    std::cout << "Done!" << std::endl;
    return 0;
}
